# -*- coding: utf-8 -*-
from __future__ import absolute_import, division, print_function

import math
from multiprocessing import Array, Value
from numbers import Number

import numpy as np
from scipy import linalg
from six import string_types
from sklearn.decomposition import PCA, IncrementalPCA
from sklearn.utils import (as_float_array, check_array, check_random_state,
                           gen_batches)
from sklearn.utils.extmath import (_incremental_mean_and_var, randomized_svd,
                                   svd_flip)
from sklearn.utils.validation import check_is_fitted

from odin.ml.base import BaseEstimator, TransformerMixin
from odin.utils import Progbar, batching, ctext, flatten_list
from odin.utils.mpi import MPI

__all__ = [
    "fast_pca",
    "MiniBatchPCA",
    "PPCA",
    "SupervisedPPCA",
]

def fast_pca(*x, n_components=None, algo='rpca', y=None,
             batch_size=1024, return_model=False,
             random_state=1234):
  """ A shortcut for many different PCA algorithms

  Parameters
  ----------
  x : {list, tuple}
    list of matrices for transformation, the first matrix will
    be used for training
  n_components : {None, int}
    number of PCA components
  algo : {'pca', 'ipca', 'ppca', 'sppca', 'plda', 'rpca'}
    different PCA algorithm:
      'ipca' - IncrementalPCA,
      'ppca' - Probabilistic PCA,
      'sppca' - Supervised Probabilistic PCA,
      'plda' - Probabilistic LDA,
      'rpca' - randomized PCA using randomized SVD
  y : {numpy.ndarray, None}
    required for labels in case of `sppca`
  batch_size : int (default: 1024)
    batch size, only used for IncrementalPCA
  return_model : bool (default: False)
    if True, return the trained PCA model as the FIRST return
  """
  batch_size = int(batch_size)
  algo = str(algo).lower()
  if algo not in ('pca', 'ipca', 'ppca', 'sppca', 'plda', 'rpca'):
    raise ValueError("`algo` must be one of the following: 'pca', "
                     "'ppca', 'plda', 'sppca', or 'rpca'; but given: '%s'" % algo)
  if algo in ('sppca', 'plda') and y is None:
    raise RuntimeError("`y` must be not None if `algo='sppca'`")
  x = flatten_list(x, level=None)
  # ====== check input ====== #
  x_train = x[0]
  x_test = x[1:]
  input_shape = None
  if x_train.ndim > 2: # only 2D for PCA
    input_shape = (-1,) + x_train.shape[1:]
    new_shape = (-1, np.prod(input_shape[1:]))
    x_train = np.reshape(x_train, new_shape)
    x_test = [np.reshape(x, new_shape) for x in x_test]
    if n_components is not None: # no need to reshape back
      input_shape = None
  # ====== train PCA ====== #
  if algo == 'sppca':
    pca = SupervisedPPCA(n_components=n_components, random_state=random_state)
    pca.fit(x_train, y)
  elif algo == 'plda':
    from odin.ml import PLDA
    pca = PLDA(n_phi=n_components, random_state=random_state)
    pca.fit(x_train, y)
  elif algo == 'pca':
    pca = PCA(n_components=n_components, random_state=random_state)
    pca.fit(x_train)
  elif algo == 'rpca':
    # we copy the implementation of RandomizedPCA because
    # it is significantly faster than PCA(svd_solver='randomize')
    pca = RandomizedPCA(n_components=n_components, iterated_power=2,
                        random_state=random_state)
    pca.fit(x_train)
  elif algo == 'ipca':
    pca = IncrementalPCA(n_components=n_components, batch_size=batch_size)
    prog = Progbar(target=x_train.shape[0],
                   print_report=False, print_summary=False, name="Fitting PCA")
    for start, end in batching(batch_size=batch_size, n=x_train.shape[0],
                               seed=1234):
      pca.partial_fit(x_train[start:end], check_input=False)
      prog.add(end - start)
  elif algo == 'ppca':
    pca = PPCA(n_components=n_components, random_state=random_state)
    pca.fit(x_train)
  # ====== transform ====== #
  x_train = pca.transform(x_train)
  x_test = [pca.transform(x) for x in x_test]
  # reshape back to original shape if necessary
  if input_shape is not None:
    x_train = np.reshape(x_train, input_shape)
    x_test = [np.reshape(x, input_shape) for x in x_test]
  # return the results
  if len(x_test) == 0:
    return x_train if not return_model else (pca, x_train)
  return tuple([x_train] + x_test) if not return_model else tuple([pca, x_train] + x_test)

# ===========================================================================
# PPCA
# ===========================================================================
class PPCA(BaseEstimator, TransformerMixin):
  """ Probabilistic Principal Components Analysis

  (C) Copyright University of Eastern Finland (UEF).
  Ville Vestman, ville.vestman@uef.fi,
  Tomi Kinnunen, tkinnu@cs.uef.fi.

  Parameters
  ----------
  n_components : {int, None}
    if None, keep the same dimensions as input features
  bias : {vector, 'auto'} [feat_dim,]
    if 'auto' take mean of training data
  n_iter : {integer, 'auto'}
    if 'auto', keep iterating until no more improvement (i.e. reduction in `sigma` value)
    compared to the `improve_threshold`
  improve_threshold : scalar
    Only used in case `n_iter='auto'`
  solver : {'traditional', 'simple'}
  verbose: {0, 1}
    showing logging information during fitting
  random_state : {None, integer, numpy.random.RandomState}

  Attributes
  ----------
  V_ : [feat_dim, n_components]
    total variability matrix
  bias_ : [feat_dim]
    bias vector
  sigma_ : scalar
    variance of error term

  References
  ----------
  [1] Ville Vestman and Tomi Kinnunen, "Supervector Compression
  Strategies to Speed up i-vector System Development",
  submitted to Speaker Odyssey 2018.

  """

  def __init__(self, n_components=None, bias='auto',
               n_iter='auto', improve_threshold=1e-3, solver='traditional',
               verbose=0, random_state=None):
    super(PPCA, self).__init__()
    if isinstance(n_components, Number):
      assert n_components > 0, \
      "`n_components` must be greater than 0, but given: %d" % n_components
      n_components = int(n_components)
    elif n_components is not None:
      raise ValueError("`n_components` can be None or integer")
    self.n_components_ = n_components
    # ====== checking bias ====== #
    if isinstance(bias, string_types):
      bias = bias.strip().lower()
      assert bias == 'auto', 'Invalid value for `bias`: %s' % bias
    elif not isinstance(bias, (np.ndarray, Number)):
      raise ValueError("`bias` can be 'auto', numpy.ndarray or a number")
    self.bias_ = bias
    # ====== checking solver ====== #
    if solver not in ('traditional', 'simple'):
      raise ValueError("`solver` must be: 'traditional', or 'simple'")
    self.solver_ = solver
    # ====== checking n_iter ====== #
    if isinstance(n_iter, string_types):
      n_iter = n_iter.lower()
      assert n_iter == 'auto', 'Invalid `n_iter` value: %s' % n_iter
    elif isinstance(n_iter, Number):
      assert n_iter > 0, "`n_iter` must greater than 0, but given: %d" % n_iter
    self.n_iter_ = n_iter
    # ====== checking random_state ====== #
    if random_state is None:
      rand = np.random.RandomState(seed=None)
    elif isinstance(random_state, Number):
      rand = np.random.RandomState(seed=None)
    elif isinstance(random_state, np.random.RandomState):
      rand = random_state
    else:
      raise ValueError("No suppport for `random_state` value: %s" % str(random_state))
    self.random_state_ = rand
    # ====== other dimension ====== #
    self.improve_threshold_ = float(improve_threshold)
    self.feat_dim_ = None
    self.verbose_ = int(verbose)

  def fit(self, X, y=None):
    # ====== initialize ====== #
    num_samples, feat_dim = X.shape
    n_components = feat_dim if self.n_components_ is None else self.n_components_
    if self.bias_ == 'auto':
      bias = np.mean(X, 0)
    elif isinstance(self.bias_, Number):
      bias = np.full(shape=(feat_dim,), fill_value=self.bias_)
    else:
      bias = self.bias_
      assert bias.shape == (feat_dim,), \
      "Invialid `bias` given shape: %s, require shape: %s" % (str(bias.shape), str((feat_dim,)))
    # ====== initialize parameters ====== #
    V = self.random_state_.rand(feat_dim, n_components)
    last_sigma = None
    sigma = 1
    centeredM = X - bias[np.newaxis, :]
    varianceM = np.sum(centeredM**2) / (num_samples * feat_dim)
    # ====== training ====== #
    if self.verbose_:
      print('[PPCA]n_components: %d  n_sample: %d  feat_dim: %d  n_iter: %d  threshold: %f  solver: %s' %
            (n_components, num_samples, feat_dim,
             -1 if self.n_iter_ == 'auto' else self.n_iter_, self.improve_threshold_, self.solver_))
    curr_n_iter = 0
    while True:
      B = (V * 1 / sigma).T # [feat_dim, n_components]
      Sigma = np.linalg.inv(np.eye(n_components) + np.dot(B, V)) # [n_components, n_components]
      my = np.dot(np.dot(Sigma, B), centeredM.T) # [n_components, num_samples]

      if self.solver_ == 'traditional':
        sumEmm = num_samples * Sigma + np.dot(my, my.T)
      elif self.solver_ == 'simple':
        sumEmm = np.dot(my, my.T)
      sumEmmInv = np.linalg.inv(sumEmm) # [n_components, n_components]
      # updating V and sigma for next iteration
      V = np.dot(np.dot(centeredM.T, my.T), sumEmmInv) # [feat_dim, n_components]
      last_sigma = sigma
      sigma = varianceM - np.sum(sumEmm * np.dot(V.T, V)) / (feat_dim * num_samples)
      improvement = last_sigma - sigma
      # log
      if self.verbose_ > 0:
        print("Iteration: %d   sigma: %.3f   improvement: %.3f" % (curr_n_iter, sigma, improvement))
      # check iteration escape
      curr_n_iter += 1
      if isinstance(self.n_iter_, Number):
        if curr_n_iter >= self.n_iter_:
          break
      elif curr_n_iter > 1 and improvement < self.improve_threshold_:
        break
    # ====== save the model ====== #
    # record new dimensions
    self.feat_dim_ = feat_dim
    self.n_components_ = n_components
    # trained vectors and matrices
    self.V_ = V
    self.bias_ = bias
    self.sigma_ = sigma
    # pre-calculate matrix for transform
    B = (V * 1 / sigma).T
    Sigma = np.linalg.inv(np.eye(n_components) + np.dot(B, V))
    self.extractorMatrix_ = np.dot(Sigma, B) # [n_components, feat_dim]

  def transform(self, X):
    """
    Parameters
    ----------
    X : matrix [num_samples, feat_dim]
    """
    assert hasattr(self, 'extractorMatrix_'), "The model hasn't `fit` on data"
    assert X.shape[1] == self.feat_dim_, \
    "Expect input matrix with shape: [?, %d], but give: %s" % (self.feat_dim_, str(X.shape))
    ivec = np.dot(self.extractorMatrix_, (X - self.bias_[np.newaxis, :]).T)
    return ivec.T

class SupervisedPPCA(PPCA):
  """ Supervised Probabilistic Principal Components Analysis

  (C) Copyright University of Eastern Finland (UEF).
  Ville Vestman, ville.vestman@uef.fi,
  Tomi Kinnunen, tkinnu@cs.uef.fi.

  Parameters
  ----------
  n_components : {int, None}
    if None, keep the same dimensions as input features
  bias : {vector, 'auto'} [feat_dim,]
    if 'auto' take mean of training data
  beta : scalar (default: 1)
    a weight parameter (use beta = 1 as default)
  n_iter : {integer, 'auto'}
    if 'auto', keep iterating until no more improvement (i.e. reduction in `sigma` value)
    compared to the `improve_threshold`
  improve_threshold : scalar
    Only used in case `n_iter='auto'`
  solver : {'traditional', 'simple'}
  extractor : {'supervised', 'unsupervised'}
    'supervised' is the probabilistic partial least squares extractor using
    both unsupervised and supervised information
  verbose: {0, 1}
    showing logging information during fitting
  random_state : {None, integer, numpy.random.RandomState}

  Attributes
  ----------
  V_ : [feat_dim, n_components]
    total variability matrix
  Q_ : [feat_dim, n_components]
    matrix for mapping speaker-dependent supervectors to i-vectors
  sigma_ : scalar
    variance of error term
  rho_ : scalar
    variance of error term in speaker-dependent supervector model
  bias_ : [feat_dim,]
    bias vector
  classBias_ : [feat_dim,]
    mean of speaker-dependent supervectors

  """

  def __init__(self, n_components=None, bias='auto', beta=1,
               n_iter='auto', improve_threshold=1e-3,
               solver='traditional', extractor='supervised',
               verbose=0, random_state=None):
    super(SupervisedPPCA, self).__init__(n_components=n_components, bias=bias,
               n_iter=n_iter, solver=solver, improve_threshold=improve_threshold,
               verbose=verbose, random_state=random_state)
    self.beta_ = float(beta)
    # ====== check extractor ====== #
    extractor = str(extractor).lower()
    if extractor not in ('supervised', 'unsupervised'):
      raise ValueError("`extractor` can only be: 'unsupervised' or 'supervised'")
    self.extractor_ = extractor

  def fit(self, X, y, z=None):
    """
    Parameters
    ----------
    X : matrix [num_samples, feat_dim]
    y : vector (int) [num_samples,]
    z : matrix [num_classes, feat_dim]
      class-dependent feature vectors for each class from 0 to `num_classes - 1`
      (in this order).
    """
    # ====== initialize ====== #
    num_samples, feat_dim = X.shape
    num_classes = z.shape[0] if z is not None else len(np.unique(y))
    n_components = feat_dim if self.n_components_ is None else self.n_components_
    if self.bias_ == 'auto':
      bias = np.mean(X, 0)
    elif isinstance(self.bias_, Number):
      bias = np.full(shape=(feat_dim,), fill_value=self.bias_)
    else:
      bias = self.bias_
      assert bias.shape == (feat_dim,), \
      "Invialid `bias` given shape: %s, require shape: %s" % (str(bias.shape), str((feat_dim,)))
    # checking `y`
    y = y.ravel().astype('int32')
    assert y.shape[0] == num_samples, \
    "Number of samples incosistent in `X`(%s) and `y`(%s)" % (str(X.shape), str(y.shape))
    # checking `z`
    if z is None:
      z = np.empty(shape=(max(np.max(y) + 1, num_classes), feat_dim),
                   dtype=X.dtype)
      for i in np.unique(y):
        z[i, :] = np.mean(X[y == i], axis=0, keepdims=True)
    else:
      assert z.shape[0] == num_classes
      assert z.shape[1] == feat_dim
    # ====== initialize parameters ====== #
    V = self.random_state_.rand(feat_dim, n_components)
    Q = self.random_state_.rand(feat_dim, n_components)
    last_sigma = None
    sigma = 1
    last_rho = None
    rho = 1

    centeredM = X - bias[np.newaxis, :]
    varianceM = np.sum(centeredM**2) / (num_samples * feat_dim)

    centeredY = z[y]
    classBias = np.mean(centeredY, 0)
    centeredY = centeredY - classBias[np.newaxis, :]
    varianceY = np.sum(centeredY**2) / (num_samples * feat_dim)
    # ====== training ====== #
    if self.verbose_:
      print('[S-PPCA]n_components: %d  n_sample: %d  feat_dim: %d  n_iter: %d  threshold: %f  solver: %s' %
            (n_components, num_samples, feat_dim,
             -1 if self.n_iter_ == 'auto' else self.n_iter_, self.improve_threshold_, self.solver_))
    curr_n_iter = 0
    while True:
      B = (V * 1 / sigma).T # [feat_dim, n_components]
      C = (Q * self.beta_ * 1 / rho).T # [feat_dim, n_components]
      Sigma = np.linalg.inv(np.eye(n_components) + np.dot(B, V) + np.dot(C, Q)) # [n_components, n_components]
      # [n_components, num_samples]
      my = np.dot(Sigma, np.dot(B, centeredM.T) + np.dot(C, centeredY.T))

      if self.solver_ == 'traditional':
        sumEmm = num_samples * Sigma + np.dot(my, my.T)
      elif self.solver_ == 'simple':
        sumEmm = np.dot(my, my.T)
      sumEmmInv = np.linalg.inv(sumEmm) # [n_components, n_components]

      # updating V and sigma for next iteration
      V = np.dot(np.dot(centeredM.T, my.T), sumEmmInv) # [feat_dim, n_components]
      Q = np.dot(np.dot(centeredY.T, my.T), sumEmmInv) # [feat_dim, n_components]

      last_sigma = sigma
      sigma = varianceM - np.sum(sumEmm * np.dot(V.T, V)) / (feat_dim * num_samples)
      improvement_sigma = last_sigma - sigma

      last_rho = rho
      rho = varianceY - np.sum(sumEmm * np.dot(Q.T, Q)) / (feat_dim * num_samples)
      improvement_rho = last_rho - rho

      # log
      if self.verbose_ > 0:
        print("Iteration: %d   sigma: %.3f   rho: %.3f    improvement: %.3f:%.3f" %
          (curr_n_iter, sigma, rho, improvement_sigma, improvement_rho))
      # check iteration escape
      curr_n_iter += 1
      if isinstance(self.n_iter_, Number):
        if curr_n_iter >= self.n_iter_:
          break
      elif curr_n_iter > 1 and \
      improvement_sigma < self.improve_threshold_ and \
      improvement_rho < self.improve_threshold_:
        break
    # ====== save the model ====== #
    # record new dimensions
    self.feat_dim_ = feat_dim
    self.n_components_ = n_components
    self.num_classes_ = num_classes
    # trained vectors and matrices
    self.V_ = V
    self.Q_ = Q
    self.bias_ = bias
    self.classBias_ = classBias
    self.sigma_ = sigma
    self.rho_ = rho
    # pre-calculate matrix for PPCA transform
    B = (V * 1 / sigma).T
    Sigma = np.linalg.inv(np.eye(n_components) + np.dot(B, V))
    self.extractorMatrix_ = np.dot(Sigma, B) # [n_components, feat_dim]
    # pre-calculate matrix for PPLS transform
    A = np.concatenate([V, Q], axis=0) # [2 * feat_dim, n_components]
    B = np.concatenate([(V * 1 / sigma).T, (Q * 1 / rho).T], axis=-1) # [n_components, 2 * feat_dim]
    sigmaW = np.linalg.inv(np.eye(n_components) + np.dot(B, A)) # [n_components, n_components]
    self.extractorMatrixPPLS_ = np.dot(sigmaW, B) # [n_components, 2 * feat_dim]

    C = np.dot(V.T, V) + sigma * np.eye(n_components) # [n_components, n_components]
    self.labelMatrix_ = np.dot(Q, np.linalg.solve(C, V.T)) # [feat_dim, feat_dim]

  def transform(self, X):
    if self.extractor_ == 'unsupervised':
      return super(SupervisedPPCA, self).transform(X)
    else:
      centeredM = X - self.bias_[np.newaxis, :]
      labels = np.dot(self.labelMatrix_, centeredM.T) + self.classBias_[:, np.newaxis]
      ivec = np.dot(self.extractorMatrixPPLS_,
                    np.concatenate([X.T, labels], axis=0) -
                    np.concatenate([self.bias_, self.classBias_])[:, np.newaxis])
      return ivec.T

# ===========================================================================
# PCA
# ===========================================================================
class RandomizedPCA(BaseEstimator, TransformerMixin):
  """Principal component analysis (PCA) using randomized SVD

  Linear dimensionality reduction using approximated Singular Value
  Decomposition of the data and keeping only the most significant
  singular vectors to project the data to a lower dimensional space.

  Parameters
  ----------
  n_components : int, optional
      Maximum number of components to keep. When not given or None, this
      is set to n_features (the second dimension of the training data).

  copy : bool
      If False, data passed to fit are overwritten and running
      fit(X).transform(X) will not yield the expected results,
      use fit_transform(X) instead.

  iterated_power : int, default=2
      Number of iterations for the power method.

  whiten : bool, optional
      When True (False by default) the `components_` vectors are multiplied
      by the square root of (n_samples) and divided by the singular values to
      ensure uncorrelated outputs with unit component-wise variances.

      Whitening will remove some information from the transformed signal
      (the relative variance scales of the components) but can sometime
      improve the predictive accuracy of the downstream estimators by
      making their data respect some hard-wired assumptions.

  random_state : int, RandomState instance or None, optional, default=None
      If int, random_state is the seed used by the random number generator;
      If RandomState instance, random_state is the random number generator;
      If None, the random number generator is the RandomState instance used
      by `np.random`.

  Attributes
  ----------
  components_ : array, shape (n_components, n_features)
      Components with maximum variance.

  explained_variance_ratio_ : array, shape (n_components,)
      Percentage of variance explained by each of the selected components.
      If k is not set then all components are stored and the sum of explained
      variances is equal to 1.0.

  singular_values_ : array, shape (n_components,)
      The singular values corresponding to each of the selected components.
      The singular values are equal to the 2-norms of the ``n_components``
      variables in the lower-dimensional space.

  mean_ : array, shape (n_features,)
      Per-feature empirical mean, estimated from the training set.

  Examples
  --------
  >>> import numpy as np
  >>> from sklearn.decomposition import RandomizedPCA
  >>> X = np.array([[-1, -1], [-2, -1], [-3, -2], [1, 1], [2, 1], [3, 2]])
  >>> pca = RandomizedPCA(n_components=2)
  >>> pca.fit(X)                 # doctest: +ELLIPSIS +NORMALIZE_WHITESPACE
  RandomizedPCA(copy=True, iterated_power=2, n_components=2,
         random_state=None, whiten=False)
  >>> print(pca.explained_variance_ratio_)  # doctest: +ELLIPSIS
  [ 0.99244...  0.00755...]
  >>> print(pca.singular_values_)  # doctest: +ELLIPSIS
  [ 6.30061...  0.54980...]

  References
  ----------

  .. [Halko2009] `Finding structure with randomness: Stochastic algorithms
    for constructing approximate matrix decompositions Halko, et al., 2009
    (arXiv:909)`

  .. [MRT] `A randomized algorithm for the decomposition of matrices
    Per-Gunnar Martinsson, Vladimir Rokhlin and Mark Tygert`

  """

  def __init__(self, n_components=None, copy=True, iterated_power=2,
               whiten=False, random_state=None):
    self.n_components = n_components
    self.copy = copy
    self.iterated_power = iterated_power
    self.whiten = whiten
    self.random_state = random_state

  def fit(self, X, y=None):
    """Fit the model with X by extracting the first principal components.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training data, where n_samples in the number of samples
        and n_features is the number of features.

    y : Ignored.

    Returns
    -------
    self : object
        Returns the instance itself.
    """
    self._fit(check_array(X))
    return self

  def _fit(self, X):
    """Fit the model to the data X.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        Training vector, where n_samples in the number of samples and
        n_features is the number of features.

    Returns
    -------
    X : ndarray, shape (n_samples, n_features)
        The input data, copied, centered and whitened when requested.
    """
    random_state = check_random_state(self.random_state)
    X = np.atleast_2d(as_float_array(X, copy=self.copy))

    n_samples = X.shape[0]

    # Center data
    self.mean_ = np.mean(X, axis=0)
    X -= self.mean_
    if self.n_components is None:
      n_components = X.shape[1]
    else:
      n_components = self.n_components

    U, S, V = randomized_svd(X, n_components,
                             n_iter=self.iterated_power,
                             random_state=random_state)

    self.explained_variance_ = exp_var = (S ** 2) / (n_samples - 1)
    full_var = np.var(X, ddof=1, axis=0).sum()
    self.explained_variance_ratio_ = exp_var / full_var
    self.singular_values_ = S  # Store the singular values.

    if self.whiten:
      self.components_ = V / S[:, np.newaxis] * math.sqrt(n_samples)
    else:
      self.components_ = V

    return X

  def transform(self, X):
    """Apply dimensionality reduction on X.

    X is projected on the first principal components previous extracted
    from a training set.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        New data, where n_samples in the number of samples
        and n_features is the number of features.

    Returns
    -------
    X_new : array-like, shape (n_samples, n_components)

    """
    check_is_fitted(self, 'mean_')

    X = check_array(X)
    if self.mean_ is not None:
      X = X - self.mean_

    X = np.dot(X, self.components_.T)
    return X

  def fit_transform(self, X, y=None):
    """Fit the model with X and apply the dimensionality reduction on X.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_features)
        New data, where n_samples in the number of samples
        and n_features is the number of features.

    y : Ignored.

    Returns
    -------
    X_new : array-like, shape (n_samples, n_components)

    """
    X = check_array(X)
    X = self._fit(X)
    return np.dot(X, self.components_.T)

  def inverse_transform(self, X):
    """Transform data back to its original space.

    Returns an array X_original whose transform would be X.

    Parameters
    ----------
    X : array-like, shape (n_samples, n_components)
        New data, where n_samples in the number of samples
        and n_components is the number of components.

    Returns
    -------
    X_original array-like, shape (n_samples, n_features)

    Notes
    -----
    If whitening is enabled, inverse_transform does not compute the
    exact inverse operation of transform.
    """
    check_is_fitted(self, 'mean_')

    X_original = np.dot(X, self.components_)
    if self.mean_ is not None:
      X_original = X_original + self.mean_
    return X_original

class MiniBatchPCA(IncrementalPCA):
  """ A modified version of IncrementalPCA to effectively
  support multi-processing (but not work)
  Original Author: Kyle Kastner <kastnerkyle@gmail.com>
                   Giorgio Patrini
  License: BSD 3 clause

  Incremental principal components analysis (IPCA).

  Linear dimensionality reduction using Singular Value Decomposition of
  centered data, keeping only the most significant singular vectors to
  project the data to a lower dimensional space.

  Depending on the size of the input data, this algorithm can be much more
  memory efficient than a PCA.

  This algorithm has constant memory complexity, on the order
  of ``batch_size``, enabling use of np.memmap files without loading the
  entire file into memory.

  The computational overhead of each SVD is
  ``O(batch_size * n_features ** 2)``, but only 2 * batch_size samples
  remain in memory at a time. There will be ``n_samples / batch_size`` SVD
  computations to get the principal components, versus 1 large SVD of
  complexity ``O(n_samples * n_features ** 2)`` for PCA.

  Read more in the :ref:`User Guide <IncrementalPCA>`.

  Parameters
  ----------
  n_components : int or None, (default=None)
      Number of components to keep. If ``n_components `` is ``None``,
      then ``n_components`` is set to ``min(n_samples, n_features)``.

  batch_size : int or None, (default=None)
      The number of samples to use for each batch. Only used when calling
      ``fit``. If ``batch_size`` is ``None``, then ``batch_size``
      is inferred from the data and set to ``5 * n_features``, to provide a
      balance between approximation accuracy and memory consumption.

  copy : bool, (default=True)
      If False, X will be overwritten. ``copy=False`` can be used to
      save memory but is unsafe for general use.

  whiten : bool, optional
      When True (False by default) the ``components_`` vectors are divided
      by ``n_samples`` times ``components_`` to ensure uncorrelated outputs
      with unit component-wise variances.

      Whitening will remove some information from the transformed signal
      (the relative variance scales of the components) but can sometimes
      improve the predictive accuracy of the downstream estimators by
      making data respect some hard-wired assumptions.

  Attributes
  ----------
  components_ : array, shape (n_components, n_features)
      Components with maximum variance.

  explained_variance_ : array, shape (n_components,)
      Variance explained by each of the selected components.

  explained_variance_ratio_ : array, shape (n_components,)
      Percentage of variance explained by each of the selected components.
      If all components are stored, the sum of explained variances is equal
      to 1.0

  mean_ : array, shape (n_features,)
      Per-feature empirical mean, aggregate over calls to ``partial_fit``.

  var_ : array, shape (n_features,)
      Per-feature empirical variance, aggregate over calls to
      ``partial_fit``.

  noise_variance_ : float
      The estimated noise covariance following the Probabilistic PCA model
      from Tipping and Bishop 1999. See "Pattern Recognition and
      Machine Learning" by C. Bishop, 12.2.1 p. 574 or
      http://www.miketipping.com/papers/met-mppca.pdf.

  n_components_ : int
      The estimated number of components. Relevant when
      ``n_components=None``.

  n_samples_seen_ : int
      The number of samples processed by the estimator. Will be reset on
      new calls to fit, but increments across ``partial_fit`` calls.

  Notes
  -----
  Implements the incremental PCA model from:
  `D. Ross, J. Lim, R. Lin, M. Yang, Incremental Learning for Robust Visual
  Tracking, International Journal of Computer Vision, Volume 77, Issue 1-3,
  pp. 125-141, May 2008.`
  See http://www.cs.toronto.edu/~dross/ivt/RossLimLinYang_ijcv.pdf

  This model is an extension of the Sequential Karhunen-Loeve Transform from:
  `A. Levy and M. Lindenbaum, Sequential Karhunen-Loeve Basis Extraction and
  its Application to Images, IEEE Transactions on Image Processing, Volume 9,
  Number 8, pp. 1371-1374, August 2000.`
  See http://www.cs.technion.ac.il/~mic/doc/skl-ip.pdf

  We have specifically abstained from an optimization used by authors of both
  papers, a QR decomposition used in specific situations to reduce the
  algorithmic complexity of the SVD. The source for this technique is
  `Matrix Computations, Third Edition, G. Holub and C. Van Loan, Chapter 5,
  section 5.4.4, pp 252-253.`. This technique has been omitted because it is
  advantageous only when decomposing a matrix with ``n_samples`` (rows)
  >= 5/3 * ``n_features`` (columns), and hurts the readability of the
  implemented algorithm. This would be a good opportunity for future
  optimization, if it is deemed necessary.

  For `multiprocessing`, you can do parallelized `partial_fit` or `transform`
  but you cannot do `partial_fit` in one process and `transform` in the others.

  Application
  -----------
  In detail, in order for PCA to work well, informally we require that
  (i) The features have approximately zero mean, and
  (ii) The different features have similar variances to each other.
  With natural images, (ii) is already satisfied even without variance
  normalization, and so we won’t perform any variance normalization.
  (If you are training on audio data—say, on spectrograms—or on text data—say,
  bag-of-word vectors—we will usually not perform variance normalization
  either.)

  By using PCA, we aim for:
  (i) the features are less correlated with each other, and
  (ii) the features all have the same variance.


  Original link: http://ufldl.stanford.edu/tutorial/unsupervised/PCAWhitening/

  References
  ----------
  D. Ross, J. Lim, R. Lin, M. Yang. Incremental Learning for Robust Visual
      Tracking, International Journal of Computer Vision, Volume 77,
      Issue 1-3, pp. 125-141, May 2008.

  G. Golub and C. Van Loan. Matrix Computations, Third Edition, Chapter 5,
      Section 5.4.4, pp. 252-253.

  See also
  --------
  PCA
  RandomizedPCA
  KernelPCA
  SparsePCA
  TruncatedSVD
  """

  def __init__(self, n_components=None, whiten=False, copy=True,
               batch_size=None):
    super(MiniBatchPCA, self).__init__(n_components=n_components,
        whiten=whiten, copy=copy, batch_size=batch_size)
    # some statistics
    self.n_samples_seen_ = 0
    self.mean_ = .0
    self.var_ = .0
    self.components_ = None
    # if nb_samples < nb_components, then the mini batch is cached until
    # we have enough samples
    self._cache_batches = []
    self._nb_cached_samples = 0

  @property
  def is_fitted(self):
    return self.components_ is not None

  # ==================== Training ==================== #
  def fit(self, X, y=None):
    """Fit the model with X, using minibatches of size batch_size.

    Parameters
    ----------
    X: array-like, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples and
        n_features is the number of features.

    y: Passthrough for ``Pipeline`` compatibility.

    Returns
    -------
    self: object
        Returns the instance itself.
    """
    X = check_array(X, copy=self.copy, dtype=[np.float64, np.float32])
    n_samples, n_features = X.shape

    if self.batch_size is None:
      batch_size = 12 * n_features
    else:
      batch_size = self.batch_size

    for batch in gen_batches(n_samples, batch_size):
      x = X[batch]
      self.partial_fit(x, check_input=False)
    return self

  def partial_fit(self, X, y=None, check_input=True):
    """Incremental fit with X. All of X is processed as a single batch.

    Parameters
    ----------
    X: array-like, shape (n_samples, n_features)
        Training data, where n_samples is the number of samples and
        n_features is the number of features.

    Returns
    -------
    self: object
        Returns the instance itself.
    """
    # ====== check the samples and cahces ====== #
    if check_input:
      X = check_array(X, copy=self.copy, dtype=[np.float64, np.float32])
    n_samples, n_features = X.shape
    # check number of components
    if self.n_components is None:
      self.n_components_ = n_features
    elif not 1 <= self.n_components <= n_features:
      raise ValueError("n_components=%r invalid for n_features=%d, need "
                       "more rows than columns for IncrementalPCA "
                       "processing" % (self.n_components, n_features))
    else:
      self.n_components_ = self.n_components
    # check the cache
    if n_samples < n_features or self._nb_cached_samples > 0:
      self._cache_batches.append(X)
      self._nb_cached_samples += n_samples
      # not enough samples yet
      if self._nb_cached_samples < n_features:
        return
      else: # group mini batch into big batch
        X = np.concatenate(self._cache_batches, axis=0)
        self._cache_batches = []
        self._nb_cached_samples = 0
    n_samples = X.shape[0]
    # ====== fit the model ====== #
    if (self.components_ is not None) and (self.components_.shape[0] !=
                                           self.n_components_):
      raise ValueError("Number of input features has changed from %i "
                       "to %i between calls to partial_fit! Try "
                       "setting n_components to a fixed value." %
                       (self.components_.shape[0], self.n_components_))
    # Update stats - they are 0 if this is the fisrt step
    col_mean, col_var, n_total_samples = \
        _incremental_mean_and_var(X, last_mean=self.mean_,
                                  last_variance=self.var_,
                                  last_sample_count=self.n_samples_seen_)
    total_var = np.sum(col_var * n_total_samples)
    if total_var == 0: # if variance == 0, make no sense to continue
      return self
    # Whitening
    if self.n_samples_seen_ == 0:
      # If it is the first step, simply whiten X
      X -= col_mean
    else:
      col_batch_mean = np.mean(X, axis=0)
      X -= col_batch_mean
      # Build matrix of combined previous basis and new data
      mean_correction = \
          np.sqrt((self.n_samples_seen_ * n_samples) /
                  n_total_samples) * (self.mean_ - col_batch_mean)
      X = np.vstack((self.singular_values_.reshape((-1, 1)) *
                    self.components_, X, mean_correction))

    U, S, V = linalg.svd(X, full_matrices=False)
    U, V = svd_flip(U, V, u_based_decision=False)
    explained_variance = S ** 2 / n_total_samples
    explained_variance_ratio = S ** 2 / total_var

    self.n_samples_seen_ = n_total_samples
    self.components_ = V[:self.n_components_]
    self.singular_values_ = S[:self.n_components_]
    self.mean_ = col_mean
    self.var_ = col_var
    self.explained_variance_ = explained_variance[:self.n_components_]
    self.explained_variance_ratio_ = \
        explained_variance_ratio[:self.n_components_]
    if self.n_components_ < n_features:
      self.noise_variance_ = \
          explained_variance[self.n_components_:].mean()
    else:
      self.noise_variance_ = 0.
    return self

  def transform(self, X, n_components=None):
    # ====== check number of components ====== #
    # specified percentage of explained variance
    if n_components is not None:
      # percentage of variances
      if n_components < 1.:
        _ = np.cumsum(self.explained_variance_ratio_)
        n_components = (_ > n_components).nonzero()[0][0] + 1
      # specific number of components
      else:
        n_components = int(n_components)
    # ====== other info ====== #
    n = X.shape[0]
    if self.batch_size is None:
      batch_size = 12 * len(self.mean_)
    else:
      batch_size = self.batch_size
    # ====== start transforming ====== #
    X_transformed = []
    for start, end in batching(n=n, batch_size=batch_size):
      x = super(MiniBatchPCA, self).transform(X=X[start:end])
      if n_components is not None:
        x = x[:, :n_components]
      X_transformed.append(x)
    return np.concatenate(X_transformed, axis=0)

  def invert_transform(self, X):
    return super(MiniBatchPCA, self).inverse_transform(X=X)

  def transform_mpi(self, X, keep_order=True, ncpu=4,
                    n_components=None):
    """ Sample as transform but using multiprocessing """
    n = X.shape[0]
    if self.batch_size is None:
      batch_size = 12 * len(self.mean_)
    else:
      batch_size = self.batch_size
    batch_list = [(i, min(i + batch_size, n))
        for i in range(0, n + batch_size, batch_size) if i < n]

    # ====== run MPI jobs ====== #
    def map_func(batch):
      start, end = batch
      x = super(MiniBatchPCA, self).transform(X=X[start:end])
      # doing dim reduction here save a lot of memory for
      # inter-processors transfer
      if n_components is not None:
        x = x[:, :n_components]
      # just need to return the start for ordering
      yield start, x
    mpi = MPI(batch_list, func=map_func,
              ncpu=ncpu, batch=1, hwm=ncpu * 12,
              backend='python')
    # ====== process the return ====== #
    X_transformed = []
    for start, x in mpi:
      X_transformed.append((start, x))
    if keep_order:
      X_transformed = sorted(X_transformed, key=lambda x: x[0])
    X_transformed = np.concatenate([x[-1] for x in X_transformed], axis=0)
    return X_transformed

  def __str__(self):
    if self.is_fitted:
      explained_vars = ';'.join([ctext('%.2f' % i, 'cyan')
                                 for i in self.explained_variance_ratio_[:8]])
    else:
      explained_vars = 0
    s = '%s(batch_size=%s, #components=%s, #samples=%s, vars=%s)' % \
        (ctext('MiniBatchPCA', 'yellow'),
         ctext(self.batch_size, 'cyan'),
         ctext(self.n_components, 'cyan'),
         ctext(self.n_samples_seen_, 'cyan'),
         explained_vars)
    return s



from __future__ import print_function, division, absolute_import

import os
import sys
import shutil

def get_script_path():
  """Return the path of the script that calling this methods"""
  path = os.path.dirname(sys.argv[0])
  path = os.path.join('.', path)
  return os.path.abspath(path)

def get_script_name():
  """Return the name of the running scipt file without extension"""
  name = os.path.basename(sys.argv[0])
  name = os.path.splitext(name)[0]
  return name



from __future__ import print_function, division, absolute_import
import matplotlib
matplotlib.use('Agg')

import os
os.environ['ODIN'] = 'float32,cpu=1,thread=1,gpu=1'
import sys
import shutil
import pickle

import numpy as np

from odin import visual as V, nnet as N
from odin.utils import ctext, unique_labels, Progbar, UnitTimer, args_parse
from odin.utils.mpi import cpu_count
from odin import fuel as F, preprocessing as pp
from utils import (PATH_ACOUSTIC, PATH_EXP, FeatureConfigs)

args = args_parse(descriptions=[
    ('--debug', 'enable debugging', None, False),
    ('-ncpu', 'if smaller than 1, auto select all possible CPU', None, 0)
])
audio = F.TIDIGITS.load()
print(audio)
all_files = sorted(list(audio['indices'].keys()))
# ===========================================================================
# Extractor
# ===========================================================================
bnf_network = N.models.BNF_2048_MFCC40()
extractors = pp.make_pipeline(steps=[
    pp.speech.AudioReader(sr=FeatureConfigs.sr, dataset=audio),
    pp.speech.PreEmphasis(coeff=0.97),
    pp.speech.Dithering(),
    # ====== STFT ====== #
    pp.speech.STFTExtractor(frame_length=FeatureConfigs.frame_length,
                            step_length=FeatureConfigs.step_length,
                            n_fft=FeatureConfigs.n_fft,
                            window=FeatureConfigs.window),
    # ====== SAD ====== #
    pp.base.RenameFeatures(input_name='stft_energy', output_name='energy'),
    pp.speech.SADgmm(nb_mixture=3, nb_train_it=25,
                           input_name='energy', output_name='sad'),
    # ====== for x-vector ====== #
    pp.speech.PowerSpecExtractor(power=2.0, output_name='spec'),
    pp.speech.MelsSpecExtractor(n_mels=24,
                                fmin=FeatureConfigs.fmin, fmax=FeatureConfigs.fmax,
                                input_name=('spec', 'sr'), output_name='mspec'),
    # ====== BNF ====== #
    pp.speech.MelsSpecExtractor(n_mels=FeatureConfigs.n_mels,
                                fmin=FeatureConfigs.fmin, fmax=FeatureConfigs.fmax,
                                input_name=('spec', 'sr'), output_name='mspec_bnf'),
    pp.speech.MFCCsExtractor(n_ceps=FeatureConfigs.n_ceps, remove_first_coef=False,
                             input_name='mspec', output_name='mfcc_bnf'),
    pp.base.AsType(dtype='float32', input_name='mfcc_bnf'),
    pp.speech.BNFExtractor(input_name='mfcc_bnf', output_name='bnf',
                           stack_context=10, pre_mvn=True,
                           sad_name='sad', remove_non_speech=False,
                           network=bnf_network,
                           batch_size=2048),
    # ====== MFCCs with deltas ====== #
    pp.speech.MFCCsExtractor(n_ceps=20, remove_first_coef=True,
                             input_name='mspec', output_name='mfcc'),
    pp.base.DeltaExtractor(input_name='mfcc', order=(0, 1, 2)),
    # ====== SDC ====== #
    pp.speech.MFCCsExtractor(n_ceps=7, remove_first_coef=True,
                             input_name='mspec', output_name='sdc'),
    pp.speech.RASTAfilter(rasta=True, input_name='sdc', output_name='sdc'),
    # ====== normalization ====== #
    pp.base.DeleteFeatures(input_name=('stft', 'spec',
                                       'mspec_bnf', 'mfcc_bnf',
                                       'sad_threshold')),
    pp.speech.AcousticNorm(mean_var_norm=True, windowed_mean_var_norm=True,
                           sad_name=None, ignore_sad_error=True,
                           input_name=('mspec', 'mfcc', 'sdc', 'bnf')),
    # ====== post processing ====== #
    pp.base.EqualizeShape0(input_name=('mspec', 'mfcc', 'sdc', 'bnf',
                                       'energy', 'sad')),
    pp.base.AsType(dtype='float16'),
], debug=args.debug)
# ====== enable debug mode ====== #
if args.debug:
  with np.warnings.catch_warnings():
    np.warnings.filterwarnings('ignore')
    for i, name in enumerate(all_files[:12]):
      tmp = extractors.transform(name)
      if isinstance(tmp, pp.base.ExtractorSignal):
        print(tmp)
        exit()
      else:
        V.plot_multiple_features(tmp, title=name)
    V.plot_save(os.path.join(PATH_EXP, 'feature_debug.pdf'))
    exit()
# ===========================================================================
# Processor
# ===========================================================================
with np.warnings.catch_warnings():
  np.warnings.filterwarnings('ignore')
  processor = pp.FeatureProcessor(jobs=all_files,
      path=PATH_ACOUSTIC,
      extractor=extractors,
      n_cache=0.12,
      ncpu =min(18, cpu_count() - 2) if args.ncpu <= 0 else int(args.ncpu),
      override=True,
      identifier='name',
      log_path=os.path.join(PATH_EXP, 'processor.log'),
      stop_on_failure=True # small dataset, enable stop on failure
  )
  with UnitTimer():
    processor.run()
  n_error = len(processor.error_log)
  print(processor)
# ====== copy readme and check the preprocessed dataset ====== #
if n_error == 0:
  readme_path = os.path.join(audio.path, [i for i in os.listdir(audio.path)
                                          if 'README' in i][0])
  shutil.copy(readme_path,
              os.path.join(PATH_ACOUSTIC, 'README.md'))

  ds = F.Dataset(PATH_ACOUSTIC, read_only=True)
  print(ds)
  pp.validate_features(ds,
                       path=os.path.join(PATH_EXP, 'acoustic'),
                       nb_samples=12, override=True)
else:
  print("%s errors happened during processing!" % ctext(n_error, 'red'))



from __future__ import absolute_import, division, print_function

import copy
import inspect
import os
import shutil
from tempfile import mkdtemp
from types import ModuleType
from typing import Callable

import dill
import tensorflow as tf
from six import string_types
from tensorflow.python import saved_model
from tensorflow.python.keras import Model
from tensorflow.python.keras import layers as layer_module
from tensorflow.python.keras.engine import base_layer_utils
from tensorflow.python.keras.layers import Layer
from tensorflow.python.util import nest


# ===========================================================================
# Helpers
# ===========================================================================
def _is_list_of_layers(obj):
  pass


# ===========================================================================
# Main
# ===========================================================================
class AdvanceModel(Model):
  """ The advance model improving the serialization and deserialization of
  complex Model

  Parameters
  ----------
  parameters : {`dict` or `None`}
    recording the arguments given to `__init__`, recommend passing
    the `locals()` dictionary
  """

  def __init__(self, parameters=None, name=None):
    super(AdvanceModel, self).__init__(name=name)
    self.supports_masking = True
    self._build_input_shape = None

    if parameters is None:
      parameters = {}
    else:
      parameters = dict(parameters)
    kwargs = parameters.pop('kwargs', {})
    parameters.update(kwargs)
    parameters.pop('self', None)
    parameters.pop('__class__', None)
    # tricky recursive reference of overriding classes
    parameters.pop('parameters', None)
    self._parameters = parameters
    self._optimizer_weights = None
    self._optimizer = None
    self._history = {}

  @property
  def train_history(self):
    history = {}
    for key, val in self._history.items():
      if 'val_' != key[:4]:
        history[key] = val
    return history

  @property
  def valid_history(self):
    history = {}
    for key, val in self._history.items():
      if 'val_' == key[:4] and key[4:] in self._history:
        history[key] = val
    return history

  @property
  def optimizer(self):
    if self._optimizer_weights is not None and \
      self._optimizer is not None and \
        len(self._optimizer.weights) == len(self._optimizer_weights):
      # TODO: this is bad solution to resume the training
      self._optimizer.set_weights(self._optimizer_weights)
      self._optimizer_weights = None
    return self._optimizer

  @optimizer.setter
  def optimizer(self, optz):
    self._optimizer = optz

  def __getstate__(self):
    configs = dill.dumps(self.get_config())
    weights = dill.dumps(self.get_weights())
    optimizer = self.optimizer
    if optimizer is not None:
      optimizer = [
          optimizer.__class__,
          optimizer.get_config(),
          optimizer.get_weights()
          if self._optimizer_weights is None else self._optimizer_weights,
      ]
      optimizer = dill.dumps(optimizer)
    return configs, weights, optimizer

  def __setstate__(self, states):
    configs, weights, optimizer = states
    configs = dill.loads(configs)
    weights = dill.loads(weights)
    clone = self.__class__.from_config(configs)
    self.__dict__.update(clone.__dict__)
    self.set_weights(weights)
    if optimizer is not None:
      cls, configs, weights = dill.loads(optimizer)
      optimizer = cls.from_config(configs)
      self.compile(optimizer)
      self._optimizer_weights = weights

  @property
  def custom_objects(self):
    """ This property could be overrided to provide custom Layer class
    Expect a dictionary mapping from class name to the class itself for
    deserialization, or, list of modules
    """
    return {}

  @property
  def parameters(self):
    return dict(self._parameters)

  @base_layer_utils.default
  def build(self, input_shape=None):
    # if default is enable, it mean the build method is not overrided
    # and won't be called during __call__
    if self._is_graph_network:
      self._init_graph_network(self.inputs, self.outputs, name=self.name)
    else:
      if input_shape is None:
        if self._build_input_shape is None:
          raise ValueError('You must provide an `input_shape` argument.')
        else:
          input_shape = self._build_input_shape
      # do not convert input_shape to tuple, multiple inputs
      # will create a ListWrapper of multiple TensorShape.
      self._build_input_shape = input_shape
      super(AdvanceModel, self).build(input_shape)
    self.built = True

  def __call__(self, inputs, *args, **kwargs):
    # This to make sure input_shapes is recorded after every call
    if self._build_input_shape is None:
      input_list = nest.flatten(inputs)
      input_shapes = None
      if all(hasattr(x, 'shape') for x in input_list):
        input_shapes = nest.map_structure(lambda x: x.shape, inputs)
      self._build_input_shape = input_shapes
    return super(AdvanceModel, self).__call__(inputs, *args, **kwargs)

  def fit(self,
          x=None,
          y=None,
          batch_size=None,
          epochs=1,
          verbose=1,
          callbacks=None,
          validation_split=0.,
          validation_data=None,
          shuffle=True,
          class_weight=None,
          sample_weight=None,
          initial_epoch=0,
          steps_per_epoch=None,
          validation_steps=None,
          validation_freq=1,
          max_queue_size=10,
          workers=1,
          use_multiprocessing=False,
          **kwargs):
    """Trains the model for a fixed number of epochs (iterations on a dataset).

    Arguments:
        x: Input data. It could be:
          - A Numpy array (or array-like), or a list of arrays
            (in case the model has multiple inputs).
          - A TensorFlow tensor, or a list of tensors
            (in case the model has multiple inputs).
          - A dict mapping input names to the corresponding array/tensors,
            if the model has named inputs.
          - A `tf.data` dataset or a dataset iterator. Should return a tuple
            of either `(inputs, targets)` or
            `(inputs, targets, sample_weights)`.
          - A generator or `keras.utils.Sequence` returning `(inputs, targets)`
            or `(inputs, targets, sample weights)`.
        y: Target data. Like the input data `x`,
          it could be either Numpy array(s) or TensorFlow tensor(s).
          It should be consistent with `x` (you cannot have Numpy inputs and
          tensor targets, or inversely). If `x` is a dataset, dataset
          iterator, generator, or `keras.utils.Sequence` instance, `y` should
          not be specified (since targets will be obtained from `x`).
        batch_size: Integer or `None`.
            Number of samples per gradient update.
            If unspecified, `batch_size` will default to 32.
            Do not specify the `batch_size` if your data is in the
            form of symbolic tensors, dataset, dataset iterators,
            generators, or `keras.utils.Sequence` instances (since they generate
            batches).
        epochs: Integer. Number of epochs to train the model.
            An epoch is an iteration over the entire `x` and `y`
            data provided.
            Note that in conjunction with `initial_epoch`,
            `epochs` is to be understood as "final epoch".
            The model is not trained for a number of iterations
            given by `epochs`, but merely until the epoch
            of index `epochs` is reached.
        verbose: 0, 1, or 2. Verbosity mode.
            0 = silent, 1 = progress bar, 2 = one line per epoch.
            Note that the progress bar is not particularly useful when
            logged to a file, so verbose=2 is recommended when not running
            interactively (eg, in a production environment).
        callbacks: List of `keras.callbacks.Callback` instances.
            List of callbacks to apply during training.
            See `tf.keras.callbacks`.
        validation_split: Float between 0 and 1.
            Fraction of the training data to be used as validation data.
            The model will set apart this fraction of the training data,
            will not train on it, and will evaluate
            the loss and any model metrics
            on this data at the end of each epoch.
            The validation data is selected from the last samples
            in the `x` and `y` data provided, before shuffling. This argument is
            not supported when `x` is a dataset, dataset iterator, generator or
           `keras.utils.Sequence` instance.
        validation_data: Data on which to evaluate
            the loss and any model metrics at the end of each epoch.
            The model will not be trained on this data.
            `validation_data` will override `validation_split`.
            `validation_data` could be:
              - tuple `(x_val, y_val)` of Numpy arrays or tensors
              - tuple `(x_val, y_val, val_sample_weights)` of Numpy arrays
              - dataset or a dataset iterator
            For the first two cases, `batch_size` must be provided.
            For the last case, `validation_steps` must be provided.
        shuffle: Boolean (whether to shuffle the training data
            before each epoch) or str (for 'batch').
            'batch' is a special option for dealing with the
            limitations of HDF5 data; it shuffles in batch-sized chunks.
            Has no effect when `steps_per_epoch` is not `None`.
        class_weight: Optional dictionary mapping class indices (integers)
            to a weight (float) value, used for weighting the loss function
            (during training only).
            This can be useful to tell the model to
            "pay more attention" to samples from
            an under-represented class.
        sample_weight: Optional Numpy array of weights for
            the training samples, used for weighting the loss function
            (during training only). You can either pass a flat (1D)
            Numpy array with the same length as the input samples
            (1:1 mapping between weights and samples),
            or in the case of temporal data,
            you can pass a 2D array with shape
            `(samples, sequence_length)`,
            to apply a different weight to every timestep of every sample.
            In this case you should make sure to specify
            `sample_weight_mode="temporal"` in `compile()`. This argument is not
            supported when `x` is a dataset, dataset iterator, generator, or
           `keras.utils.Sequence` instance, instead provide the sample_weights
            as the third element of `x`.
        initial_epoch: Integer.
            Epoch at which to start training
            (useful for resuming a previous training run).
        steps_per_epoch: Integer or `None`.
            Total number of steps (batches of samples)
            before declaring one epoch finished and starting the
            next epoch. When training with input tensors such as
            TensorFlow data tensors, the default `None` is equal to
            the number of samples in your dataset divided by
            the batch size, or 1 if that cannot be determined. If x is a
            `tf.data` dataset or a dataset iterator, and 'steps_per_epoch'
            is None, the epoch will run until the input dataset is exhausted.
        validation_steps: Only relevant if `validation_data` is provided and
            is a dataset or dataset iterator. Total number of steps (batches of
            samples) to draw before stopping when performing validation
            at the end of every epoch. If validation_data is a `tf.data` dataset
            or a dataset iterator, and 'validation_steps' is None, validation
            will run until the `validation_data` dataset is exhausted.
        validation_freq: Only relevant if validation data is provided. Integer
            or `collections.Container` instance (e.g. list, tuple, etc.). If an
            integer, specifies how many training epochs to run before a new
            validation run is performed, e.g. `validation_freq=2` runs
            validation every 2 epochs. If a Container, specifies the epochs on
            which to run validation, e.g. `validation_freq=[1, 2, 10]` runs
            validation at the end of the 1st, 2nd, and 10th epochs.
        max_queue_size: Integer. Used for generator or `keras.utils.Sequence`
            input only. Maximum size for the generator queue.
            If unspecified, `max_queue_size` will default to 10.
        workers: Integer. Used for generator or `keras.utils.Sequence` input
            only. Maximum number of processes to spin up
            when using process-based threading. If unspecified, `workers`
            will default to 1. If 0, will execute the generator on the main
            thread.
        use_multiprocessing: Boolean. Used for generator or
            `keras.utils.Sequence` input only. If `True`, use process-based
            threading. If unspecified, `use_multiprocessing` will default to
            `False`. Note that because this implementation relies on
            multiprocessing, you should not pass non-picklable arguments to
            the generator as they can't be passed easily to children processes.
        **kwargs: Used for backwards compatibility.

    Returns:
        A `History` object. Its `History.history` attribute is
        a record of training loss values and metrics values
        at successive epochs, as well as validation loss values
        and validation metrics values (if applicable).

    Raises:
        RuntimeError: If the model was never compiled.
        ValueError: In case of mismatch between the provided input data
            and what the model expects.
    """
    history = super(AdvanceModel,
                    self).fit(x=x,
                              y=y,
                              batch_size=batch_size,
                              epochs=epochs,
                              verbose=verbose,
                              callbacks=callbacks,
                              validation_split=validation_split,
                              validation_data=validation_data,
                              shuffle=shuffle,
                              class_weight=class_weight,
                              sample_weight=sample_weight,
                              initial_epoch=initial_epoch,
                              steps_per_epoch=steps_per_epoch,
                              validation_steps=validation_steps,
                              validation_freq=validation_freq,
                              max_queue_size=max_queue_size,
                              workers=workers,
                              use_multiprocessing=use_multiprocessing,
                              **kwargs)
    for key, val in history.history.items():
      if key not in self._history:
        self._history[key] = []
      self._history[key] += list(val)
    return history

  def get_config(self):
    subclass_types = []
    for t in type.mro(type(self)):
      if t == AdvanceModel:
        break
      subclass_types.append(t)
    default_types = [t for t in type.mro(type(self)) if t not in subclass_types]

    default_keys = []
    default_source = ''
    for t in default_types:
      default_keys += dir(t)
      # some class is builtin, so impossible to get the source code
      try:
        default_source += inspect.getsource(t)
      except:
        pass

    source = ''
    for t in subclass_types:
      source += inspect.getsource(t)

    attributes = {}
    layer_attributes = {}  # mapping attribute_name -> Layer
    for key, val in self.__dict__.items():
      # TODO: not a good solution here, history make recursive reference
      # but couldn't found in Model soruce code where it is setted
      if key in ('_parameters', 'history') or key in self.parameters:
        continue
      if isinstance(val, Layer):
        layer_attributes[id(val)] = key
      elif isinstance(val, tf.Variable):
        pass
      elif key not in default_keys and \
        'self.' + key not in default_source and \
          'self.' + key in source:
        try:
          attr = getattr(self, key)
          if not inspect.ismethod(attr) and \
            not isinstance(attr, property) and \
              not isinstance(attr, classmethod):
            attributes[key] = val
        except AttributeError:
          pass

    layer_configs = []
    for layer in self.layers:
      layer_configs.append({
          'class_name': layer.__class__.__name__,
          'config': layer.get_config(),
          'attribute': layer_attributes.get(id(layer), None)
      })

    config = {
        'name': self.name,
        'layers': copy.deepcopy(layer_configs),
        'parameters': self.parameters,
        'build_input_shape': self._build_input_shape,
        'attributes': attributes,
        'history': self._history,
    }
    return config

  @classmethod
  def from_config(cls, config, custom_objects=None):
    if 'name' in config:
      name = config['name']
      build_input_shape = config['build_input_shape']
      layer_configs = config['layers']
      parameters = config['parameters']
      attributes = config['attributes']
      history = config['history']
    else:
      name = None
      build_input_shape = None
      layer_configs = config
      parameters = {}
      attributes = {}
      history = {}
    # create new instance
    if 'name' in inspect.getfullargspec(cls.__init__).args:
      parameters['name'] = name
    model = cls(**parameters)
    model._history = history
    # set all the attributes
    for key, val in attributes.items():
      setattr(model, key, val)
    # preprocessing the custom_objects
    if custom_objects is None:
      custom_objects = {}
    if hasattr(model, 'custom_objects'):
      model_objects = model.custom_objects
      if isinstance(model_objects, (tuple, list)):
        for obj in model_objects:
          if isinstance(obj, type):
            custom_objects[obj.__name__] = obj
          elif isinstance(obj, ModuleType):
            for i, j in inspect.getmembers(obj):
              if isinstance(j, type) and issubclass(j, Layer):
                custom_objects[i] = j
          elif inspect.isfunction(obj) or inspect.ismethod(obj):
            custom_objects[obj.__name__] = obj
          else:
            raise ValueError(
                "Cannot process value with type %s for custom_objects" +
                " (module, type or callable are supported)" % str(type(obj)))
      elif isinstance(model_objects, dict):
        custom_objects.update(model_objects)
      else:
        raise ValueError(
            "Class %s should return custom_objects of type dictionary or list,"
            " but the returned value is %s" %
            (str(cls), str(type(model_objects))))
    # deserialize all layers (NOTE: this no created layer is assigned
    # to our model, just to check all serialization go OK)
    layers = []
    for layer_config in layer_configs:
      attr = layer_config.pop('attribute')
      try:
        layer = layer_module.deserialize(layer_config,
                                         custom_objects=custom_objects)
      except ValueError:
        layer = None
      layers.append((attr, layer))
    # build if necessary
    if not model.inputs and build_input_shape is not None:
      model.build(build_input_shape)
    # check if all layers is deserialized, if any Layer is missing
    # that mean the Layer is externally added later after build,
    # then we add it back again to the Model (this logic might be fault)
    if len(model.layers) != len(layers):
      raise RuntimeError("No support for this case")
    return model



# ===========================================================================
# Using TIDIGITS dataset to predict gender (Boy, Girl, Woman, Man)
# ===========================================================================
# Saved WAV file format:
#     0) [train|test]
#     1) [m|w|b|g] (alias for man, women, boy, girl)
#     2) [age]
#     3) [dialectID]
#     4) [speakerID]
#     5) [production]
#     6) [digit_sequence]
#     => "train_g_08_17_as_a_4291815"
# ===========================================================================
from __future__ import print_function, division, absolute_import

import os
import shutil
from collections import defaultdict

import numpy as np

from odin import fuel as F, backend as K, visual as V
from odin.utils import (cache_disk, ctext, unique_labels,
                        Progbar, batching, get_exppath,
                        get_formatted_datetime)
from odin.stats import train_valid_test_split, sampling_iter

_support_label = {
    'other': 0,
    'gender': 1,
    'age': 2,
    'dialect': 3,
    'speaker': 4,
    'production': 5,
    'digit': 6,
}

# ===========================================================================
# Const for path and features configuration
# ===========================================================================
PATH_EXP = get_exppath(tag='TIDIGITS', override=False)
# ====== acoustic feature extraction ====== #
PATH_ACOUSTIC = os.path.join(PATH_EXP, 'TIDIGITS_feat')

class FeatureConfigs(object):
  padding = False
  sr = 8000
  window = 'hamm'
  n_fft = 512
  n_mels = 40
  n_ceps = 40
  fmin = 100
  fmax = 4000
  frame_length = 0.025
  step_length = 0.010
  dtype = 'float16'

def get_exp_path(system_name, args, override=False):
  """ Return: exp_dir, model_path, log_path """
  exp_dir = get_exppath(tag='TIDIGITS_%s_%s_%s' %
    (system_name, args.task, args.feat))
  if 'nmix' in args:
    exp_dir += '_%d' % args.nmix
  if 'tdim' in args:
    exp_dir += '_%d' % args.tdim
  # ====== check override ====== #
  if bool(override) and os.path.exists(exp_dir):
    shutil.rmtree(exp_dir)
  if not os.path.exists(exp_dir):
    os.mkdir(exp_dir)
  # ====== basic paths ====== #
  model_path = os.path.join(exp_dir, 'model.ai')
  log_path = os.path.join(exp_dir,
                         'log_%s.txt' % get_formatted_datetime(only_number=True))
  print("Exp dir:", ctext(exp_dir, 'cyan'))
  print("Model path:", ctext(model_path, 'cyan'))
  print("Log path:", ctext(log_path, 'cyan'))
  return exp_dir, model_path, log_path

# ===========================================================================
# For DNN
# ===========================================================================
def prepare_data(feat, label, utt_length=0.4, for_ivec=False):
  """

  Returns (i-vector)
  ------------------
  ds[feat]
  train_files
  y_train
  test_files
  y_test
  labels

  Returns (x-vector)
  ------------------
  train : Feeder
    feeder for training data for iterating over pair of (X, y)
  valid : Feeder
    feeder for validating data for iterating over pair of (X, y)
  X_test_name : list of file names
    file names are append with '.%d' for cut segment ID
  X_test_true : list of integer
    label of each sample
  X_test_data : array
    list of test data same length as X_test_name
  labels : list of string
    list of labels for classification task

  Example
  -------
  (train, valid,
   X_test_name, X_test_true, X_test_data,
   labels) = prepare_data_dnn(feat=FEAT, label='gender')

  """
  label = str(label).lower()
  assert label in _support_label, "No support for label: %s" % label
  assert 0 < utt_length <= 1.
  # ====== load dataset ====== #
  if not os.path.exists(PATH_ACOUSTIC):
    raise RuntimeError("Cannot find extracted acoustic features at path: '%s',"
                       "run the code speech_features_extraction.py!" % PATH_ACOUSTIC)
  ds = F.Dataset(PATH_ACOUSTIC, read_only=True)
  assert feat in ds, "Cannot find feature with name: %s" % feat
  indices = list(ds['indices'].items())
  K.get_rng().shuffle(indices)

  # ====== helper ====== #
  def is_train(x):
    return x.split('_')[0] == 'train'

  def extract_label(x):
    return x.split('_')[_support_label[label]]

  print("Task:", ctext(label, 'cyan'))
  fn_label, labels = unique_labels([i[0] for i in indices],
                                   key_func=extract_label,
                                   return_labels=True)
  print("Labels:", ctext(labels, 'cyan'))
  # ====== training and test data ====== #
  train_files = [] # (name, (start, end)) ...
  test_files = []
  for name, (start, end) in indices:
    if is_train(name):
      train_files.append((name, (start, end)))
    else:
      test_files.append((name, (start, end)))
  # name for each dataset, useful for later
  print("#Train:", ctext(len(train_files), 'cyan'))
  print("#Test:", ctext(len(test_files), 'cyan'))
  # ====== for i-vectors ====== #
  y_train = np.array([fn_label(i[0]) for i in train_files])
  y_test = np.array([fn_label(i[0]) for i in test_files])
  if bool(for_ivec):
    return ds[feat], train_files, y_train, test_files, y_test, labels
  # ====== length ====== #
  length = [(end - start) for _, (start, end) in indices]
  max_length = max(length)
  frame_length = int(max_length * utt_length)
  step_length = frame_length
  print("Max length  :", ctext(max_length, 'yellow'))
  print("Frame length:", ctext(frame_length, 'yellow'))
  print("Step length :", ctext(step_length, 'yellow'))
  # ====== split dataset ====== #
  # split by speaker ID
  train_files, valid_files = train_valid_test_split(
      x=train_files, train=0.8,
      cluster_func=None,
      idfunc=lambda x: x[0].split('_')[4], # splited by speaker
      inc_test=False)
  print("#File train:", ctext(len(train_files), 'cyan'))
  print("#File valid:", ctext(len(valid_files), 'cyan'))
  print("#File test :", ctext(len(test_files), 'cyan'))

  recipes = [
      F.recipes.Sequencing(frame_length=frame_length, step_length=step_length,
                           end='pad', pad_mode='post', pad_value=0),
      F.recipes.Name2Label(converter_func=fn_label),
      F.recipes.LabelOneHot(nb_classes=len(labels), data_idx=-1)
  ]
  feeder_train = F.Feeder(F.IndexedData(ds[feat], indices=train_files),
                          ncpu=6, batch_mode='batch')
  feeder_valid = F.Feeder(F.IndexedData(ds[feat], indices=valid_files),
                          ncpu=4, batch_mode='batch')
  feeder_test = F.Feeder(F.IndexedData(ds[feat], indices=test_files),
                         ncpu=4, batch_mode='file')
  feeder_train.set_recipes(recipes)
  feeder_valid.set_recipes(recipes)
  feeder_test.set_recipes(recipes)
  print(feeder_train)

  # ====== process X_test, y_test in advance for faster evaluation ====== #
  @cache_disk
  def _extract_test_data(feat, label, utt_length):
    prog = Progbar(target=len(feeder_test),
                   print_summary=True, name="Preprocessing test set")
    X_test = defaultdict(list)
    for name, idx, X, y in feeder_test:
      # validate everything as expected
      assert fn_label(name) == np.argmax(y), name # label is right
      # save to list
      X_test[name].append((idx, X))
      prog.add(X.shape[0])
    # ====== create 1 array for data and dictionary for indices ====== #
    X_test_name = []
    X_test_data = []
    for name, X in X_test.items():
      X = np.concatenate([x[1] for x in sorted(X, key=lambda i: i[0])],
                         axis=0).astype('float16')
      X_test_name += [name + '.%d' % i for i in range(len(X))]
      X_test_data.append(X)
    X_test_name = np.array(X_test_name)
    X_test_data = np.concatenate(X_test_data, axis=0)
    return X_test_name, X_test_data
  # convert everything back to float32
  X_test_name, X_test_data = _extract_test_data(feat, label, utt_length)
  X_test_true = np.array([fn_label(i.split('.')[0])
                          for i in X_test_name])
  return feeder_train, feeder_valid, \
  X_test_name, X_test_true, X_test_data, labels

def make_dnn_prediction(functions, X, batch_size=256, title=''):
  return_list = True
  if not isinstance(functions, (tuple, list)):
    functions = [functions]
    return_list = False
  n_functions = len(functions)
  results = [[] for i in range(n_functions)]
  # ====== prepare progress bar ====== #
  n_samples = len(X)
  prog = Progbar(target=n_samples, print_summary=True,
                 name="Making prediction: %s" % str(title))
  # ====== for feeder ====== #
  if isinstance(X, F.Feeder):
    y_true = []
    for x, y in X.set_batch(batch_size=batch_size):
      for res, fn in zip(results, functions):
        res.append(fn(x))
      prog.add(x.shape[0])
      y_true.append(np.argmax(y, axis=-1) if y.ndim == 2 else y)
    results = [np.concatenate(res, axis=0)
               for res in results]
    y_true = np.concatenate(y_true, axis=0)
    if return_list:
      return results, y_true
    return results[0], y_true
  # ====== for numpy array ====== #
  else:
    for start, end in batching(batch_size=batch_size, n=n_samples):
      y = X[start:end]
      for res, fn in zip(results, functions):
        res.append(fn(y))
      prog.add(end - start)
    results = [np.concatenate(res, axis=0)
               for res in results]
    if return_list:
      return results
    return results[0]

# ===========================================================================
# visualization
# ===========================================================================
def visualize_latent_space(X_org, X_latent, name, labels, title):
  """
  X_org : [n_samples, n_timesteps, n_features]
  X_latent : [n_samples, n_timesteps, n_latents]
  """
  assert X_org.shape[0] == X_latent.shape[0] == len(name) == len(labels)
  assert not np.any(np.isnan(X_org))
  assert not np.any(np.isnan(X_latent))
  X_org = X_org.astype('float32')
  X_latent = X_latent.astype('float32')
  # ====== evaluation of the latent space ====== #
  n_channels = 1 if X_latent.ndim == 3 else int(np.prod(X_latent.shape[3:]))
  n_samples = X_org.shape[0]
  # 1 for original, 1 for mean channel, then the rest
  n_row = 1 + 1 + n_channels
  n_col = 3
  V.plot_figure(nrow=n_row + 1, ncol=16)
  # only select 3 random sample
  for i, idx in enumerate(
      sampling_iter(it=range(n_samples), k= n_col, seed=1234)):
    x = X_org[idx]
    # latent tensor can be 3D or 4D
    z = X_latent[idx]
    if z.ndim > 3:
      z = np.reshape(z, newshape=(z.shape[0], z.shape[1], -1))
    elif z.ndim == 2:
      z = np.reshape(z, newshape=(z.shape[0], z.shape[1], 1))
    elif z.ndim == 3:
      pass
    else:
      raise ValueError("No support for z value: %s" % str(z.shape))
    # plot original acoustic
    ax = V.plot_spectrogram(x.T, ax=(n_row, n_col, i + 1), title='Org')
    if i == 0:
      ax.set_title("[%s]'%s-%s'" % (str(title), str(name[idx]), str(labels[idx])),
                   fontsize=8)
    else:
      ax.set_title("'%s-%s'" % (str(name[idx]), str(labels[idx])),
                   fontsize=8)
    # plot the mean
    V.plot_spectrogram(np.mean(z, axis=-1).T,
                       ax=(n_row, n_col, i + 4), title='Zmean')
    # plot first 25 channels
    if n_channels > 1:
      for j in range(min(8, n_channels)):
        V.plot_spectrogram(z[:, :, j].T,
                           ax=(n_row, n_col, j * 3 + 7 + i),
                           title='Z%d' % j)



from __future__ import print_function, division, absolute_import

import os
import subprocess
from io import BytesIO
from six import string_types
from collections import defaultdict

import numpy as np
import soundfile as sf

from odin import nnet as N
from odin import fuel as F, preprocessing as pp

from helpers import Config, PATH_ACOUSTIC_FEATURES, CURRENT_STATE, SystemStates
# ===========================================================================
# Customized Audio Reader Extractor
# `row`:
#  0       1      2      3       4          5         6
# path, channel, name, spkid, dataset, start_time, end_time
# ===========================================================================
class SREAudioReader(pp.base.Extractor):
  """ SREAudioReader """

  def __init__(self):
    super(SREAudioReader, self).__init__(is_input_layer=True)

  def _transform(self, row):
    if len(row) == 7:
      path, channel, name, spkid, dataset, start_time, end_time = row
    else:
      path, channel, name, spkid, dataset = row[:5]
      start_time = None
      end_time = None
    # ====== read audio ====== #
    # for voxceleb1
    if dataset == 'voxceleb1':
      with open(path, 'rb') as f:
        y, sr = sf.read(f)
        y = pp.signal.resample(y, sr_orig=sr, sr_new=8000,
                               best_algorithm=True)
        sr = 8000
    # for sre, fisher and swb
    elif (dataset[:3] == 'sre' or
     dataset == 'swb' or
     dataset == 'fisher'):
      with open(path, 'rb') as f:
        y, sr = sf.read(f)
        y = pp.signal.resample(y, sr_orig=sr, sr_new=8000,
                               best_algorithm=True)
        if y.ndim == 2:
          y = y[:, int(channel)]
        sr = 8000
    # all other dataset: mix6, voxceleb2
    else:
      y, sr = pp.signal.anything2wav(inpath=path, outpath=None,
                                     channel=channel,
                                     dataset=dataset,
                                     start=start_time, end=end_time,
                                     sample_rate=Config.SAMPLE_RATE,
                                     return_data=True)
    # ====== error happen ignore file ====== #
    if len(y) == 0:
      return None
    # ====== remove DC offset ====== #
    y = y - np.mean(y, 0)
    duration = max(y.shape) / sr
    ret = {'raw': y, 'sr': sr, 'duration': duration, # in second
           'path': path,
           'spkid': spkid,
           'name': name,
           'dsname': dataset}
    return ret

class SREAugmentor(pp.base.Extractor):
  """ SREAugmentor
  New name for each utterance is:
    [utt_name]/[noise1_name]/[noise2_name]...
  """

  def __init__(self, noise_ds):
    super(SREAugmentor, self).__init__(is_input_layer=True)
    from helpers import ALL_NOISE
    self.noise_ds = str(noise_ds)
    assert self.noise_ds in ALL_NOISE, \
    "Cannot find noise dataset with name: %s; given following option: %s" % \
    (self.noise_ds, ', '.join(list(ALL_NOISE.keys())))
    # ====== mapping noise_type -> list of row ====== #
    self.noise_type = defaultdict(list)
    for path, channel, name, ntype, duration in ALL_NOISE[self.noise_ds]:
      self.noise_type[ntype].append((path, name, duration))
    self.noise_type = {k: np.array(v)
                       for k, v in self.noise_type.items()}
    # ====== fixed RandomState ====== #
    self.rand = np.random.RandomState(seed=Config.SUPER_SEED)

  def _transform(self, row):
    # we need to know the audio duration here
    path, channel, name, spkid, dataset, start_time, end_time, duration = row
    duration = float(duration)
    # rirs: mediumroom, smallroom
    # musan: music, speech, noise
    noise_type = self.rand.choice(a=sorted((self.noise_type.keys())),
                                  size=1, replace=False)[0]
    noise_data = self.noise_type[noise_type]
    # ====== wav command ====== #
    cmd_wav = pp.signal.anything2wav(inpath=path, outpath=None,
                                     channel=channel,
                                     dataset=dataset,
                                     start=start_time, end=end_time,
                                     sample_rate=Config.SAMPLE_RATE,
                                     codec='pcm16',
                                     return_data=False)
    # ====== reverberation ====== #
    if self.noise_ds == 'rirs':
      idx = self.rand.randint(low=0, high=len(noise_data), size=1, dtype=int)[0]
      noise_path, noise_name, noise_dur = noise_data[idx]
      cmd = '%s | ' + \
      'wav-reverberate --shift-output=true --impulse-response="sox %s -r 8000 -t wav - |"  - -'
      cmd = cmd % (cmd_wav, noise_path)
      # update the name
      name += '/%s' % noise_name
    # ====== MUSAN ====== #
    elif self.noise_ds == 'musan':
      ### noise (in kaldi: noise snrs is choose from one of
      # following value 15:10:5:0)
      if noise_type == 'noise':
        # duration until start the next noise audio
        noise_interval = 1 # in second
        noise = []; noise_start = []; noise_snrs = []
        curr_dur = 0; indices = np.arange(len(noise_data))
        # adding noise until the end of the utterance
        while curr_dur < duration:
          idx = self.rand.choice(indices, size=1)[0]
          noise_path, noise_name, noise_dur = noise_data[idx]
          # this does not apply anymore if we don't have enough
          # noise duration
          if len(noise) < len(noise_data) and noise_path in noise:
            continue
          noise.append(noise_path)
          noise_start.append(curr_dur)
          noise_snrs.append(self.rand.randint(low=0, high=15, size=1, dtype=int)[0])
          curr_dur += float(noise_dur) + noise_interval
          # update the name
          name += '/%s' % noise_name
        # start creating command, careful all space must be there
        cmd = '%s | wav-reverberate --shift-output=true --additive-signals=\'' % cmd_wav
        for path in noise:
          cmd += 'sox -t wav %s -r 8k -t wav - |,' % path
        cmd = cmd[:-1]
        cmd += '\' --start-times=\'%s\' ' % ','.join(['%f' % i for i in noise_start])
        cmd += '--snrs=\'%s\' - -' % ','.join(['%d' % i for i in noise_snrs])
      ### music in background
      elif noise_type == 'music':
        idx = self.rand.randint(low=0, high=len(noise_data), size=1, dtype=int)[0]
        noise_path, noise_name, noise_dur = noise_data[idx]
        snrs = self.rand.randint(low=5, high=15, size=1, dtype=int)[0]
        cmd = '%s | ' + \
        'wav-reverberate --shift-output=true --additive-signals=' + \
        '\'wav-reverberate --duration=%f "sox -t wav %s -r 8k -t wav - |" - |\' ' + \
        '--start-times=\'0\' --snrs=\'%d\' - -'
        cmd = cmd % (cmd_wav, duration, noise_path, snrs)
        # update the name
        name += '/%s' % noise_name
      ### combined multiple speech from different speakers
      #(in kaldi: snrs is one of 20:17:15:13)
      elif noise_type == 'speech':
        n_speaker = self.rand.randint(3, 7, size=1, dtype=int)[0]
        idx = self.rand.choice(a=np.arange(len(noise_data)),
                               size=n_speaker, replace=False)
        speech = noise_data[idx]
        snrs = self.rand.randint(low=13, high=20, size=n_speaker, dtype=int)
        # start creating command, careful all space must be there
        cmd = '%s | wav-reverberate --shift-output=true --additive-signals=\'' % cmd_wav
        for spk_path, spk_name, spk_dur in speech:
          cmd += 'wav-reverberate --duration=%f "sox -t wav %s -r 8k -t wav - |" - |,' % \
          (duration, spk_path)
          # update the name
          name += '/%s' % spk_name
        cmd = cmd[:-1] # remove the `,` in the end
        cmd += '\' --start-times=\'%s\' ' % ','.join(['0'] * n_speaker)
        cmd += '--snrs=\'%s\' - -' % ','.join(['%d' % i for i in snrs])
      ### Error
      else:
        raise RuntimeError("Unknown MUSAN noise type: %s" % noise_type)
    # ====== error ====== #
    else:
      raise RuntimeError("No support noise dataset with name: %s" % self.noise_ds)
    # ====== get the data ====== #
    try:
      with subprocess.Popen(args=cmd, shell=True,
                           stdout=subprocess.PIPE,
                           stderr=subprocess.PIPE) as p:
        data = BytesIO(p.stdout.read())
        y, sr = sf.read(data)
    except Exception as e:
      signal = pp.base.ExtractorSignal()
      signal.set_message(extractor=self,
                         msg=str(e),
                         last_input=row)
      signal.set_action(self.action)
      return signal
    # ====== error happen ignore file ====== #
    if len(y) == 0:
      return None
    # ====== remove DC offset ====== #
    y = y - np.mean(y, 0)
    duration = max(y.shape) / sr
    return {'raw': y, 'sr': sr, 'duration': duration, # in second
            'path': path, 'spkid': spkid, 'name': name,
            'dsname': dataset,
            'dsnoise': self.noise_ds,
            'noisetype': noise_type,
            'cmd': cmd}

class SADreader(pp.base.Extractor):
  """ This class read SAD label from acoustic dataset and
  use it for augmentation dataset
  """

  def __init__(self, ds_path):
    super(SADreader, self).__init__()
    indices = os.path.join(ds_path, 'indices_sad')
    assert os.path.isfile(indices), "Cannot find indices at path: %s" % indices
    self.indices = F.MmapDict(indices, read_only=True)

    data = os.path.join(ds_path, 'sad')
    assert os.path.isfile(data), "Cannot find SAD at path: %s" % data
    self.data = F.MmapData(data, read_only=True)

  def _transform(self, X):
    name = X['name'].split('/')[0]
    if name not in self.indices:
      return None
    start, end = self.indices[name]
    sad = self.data[start:end][:]
    return {'sad': sad}

# ===========================================================================
# Extractor
# NOTE: you must save the SAD label for augmentation data later
# ===========================================================================
def mfcc(augmentation=None):
  delete_list = ['stft', 'spec', 'raw',
                 'mfcc_energy', 'sad_threshold']
  if augmentation is not None:
    delete_list.append('sad')

  extractors = pp.make_pipeline(steps=[
      SREAugmentor(augmentation)
      if isinstance(augmentation, string_types) else
      SREAudioReader(),
      pp.speech.PreEmphasis(coeff=0.97, input_name='raw'),
      # ====== STFT ====== #
      pp.speech.STFTExtractor(frame_length=Config.FRAME_LENGTH,
                              step_length=Config.STEP_LENGTH,
                              n_fft=Config.NFFT, window=Config.WINDOW,
                              padding=False, energy=False),
      # ====== for x-vector ====== #
      pp.speech.PowerSpecExtractor(power=2.0,
                                   input_name='stft', output_name='spec'),
      pp.speech.MelsSpecExtractor(n_mels=24, fmin=20, fmax=3700,
                                  input_name=('spec', 'sr'), output_name='mspec'),
      pp.speech.MFCCsExtractor(n_ceps=24,
                               remove_first_coef=True, first_coef_energy=True,
                               input_name='mspec', output_name='mfcc'),
      # ====== extract SAD ====== #
      pp.speech.SADthreshold(energy_threshold=0.5, energy_mean_scale=0.5,
                             frame_context=2, proportion_threshold=0.12,
                             smooth_window=5,
                             input_name='mfcc_energy', output_name='sad')
      if augmentation is None else
      SADreader(ds_path=os.path.join(PATH_ACOUSTIC_FEATURES, 'mfcc')),
      pp.speech.ApplyingSAD(input_name=('mspec', 'mfcc'), sad_name='sad',
          keep_unvoiced=False if CURRENT_STATE == SystemStates.EXTRACT_FEATURES else True),
      # ====== normalization ====== #
      pp.speech.AcousticNorm(mean_var_norm=True, windowed_mean_var_norm=True,
                             win_length=301, input_name=('mspec', 'mfcc')),
      # ====== post processing ====== #
      pp.base.DeleteFeatures(input_name=delete_list),
      pp.base.AsType(dtype='float16'),
  ])
  return extractors

def bnf(augmentation=None):
  raise NotImplementedError
  bnf_network = N.models.BNF_2048_MFCC40()
  recipe = pp.make_pipeline(steps=[
      SREAudioReader(),
      pp.speech.PreEmphasis(coeff=0.97),
      # ====== STFT ====== #
      pp.speech.STFTExtractor(frame_length=Config.FRAME_LENGTH,
                              step_length=Config.STEP_LENGTH,
                              n_fft=Config.NFFT,
                              window=Config.WINDOW),
      # ====== SAD ====== #
      pp.base.RenameFeatures(input_name='stft_energy', output_name='energy'),
      pp.speech.SADgmm(nb_mixture=3, smooth_window=3,
                       input_name='energy', output_name='sad'),
      # ====== BNF ====== #
      pp.speech.PowerSpecExtractor(power=2.0, output_name='spec'),
      pp.speech.MelsSpecExtractor(n_mels=Config.NCEPS,
                                  fmin=Config.FMIN, fmax=Config.FMAX,
                                  input_name='spec', output_name='mspec'),
      pp.speech.MFCCsExtractor(n_ceps=Config.NCEPS, remove_first_coef=False,
                               input_name='mspec', output_name='mfcc'),
      pp.base.AsType(dtype='float32', input_name='mfcc'),
      pp.speech.BNFExtractor(input_name='mfcc', output_name='bnf', sad_name='sad',
                             network=bnf_network,
                             remove_non_speech=True,
                             stack_context=10, pre_mvn=True,
                             batch_size=1234),
      # ====== normalization ====== #
      pp.speech.AcousticNorm(input_name=('bnf',),
                             mean_var_norm=True,
                             windowed_mean_var_norm=True,
                             win_length=301),
      # ====== cleaning ====== #
      pp.base.DeleteFeatures(input_name=('stft', 'raw', 'energy',
                                         'sad_threshold',
                                         'spec', 'mspec', 'mfcc')),
      pp.base.AsType(dtype='float16')
  ])
  return recipe



# -*- coding: utf-8 -*-
# ===========================================================================
# Copyright 2016-2017 TrungNT
# ===========================================================================
from __future__ import print_function, division, absolute_import

import os
import sys
import inspect
import marshal
import warnings
from array import array
from six.moves import builtins

from collections import OrderedDict, defaultdict
from collections import MutableMapping, Mapping
from functools import wraps, partial
from six import string_types
from six.moves import zip, zip_longest, cPickle
import types

import numpy as np

__all__ = [
    'typecheck',
    'autoattr',
    'abstractstatic',
    'functionable',
    'singleton'
]


# ===========================================================================
# Type enforcement
# ===========================================================================
def _info(fname, expected, actual, flag):
  '''Convenience function outputs nicely formatted error/warning msg.'''
  def to_str(t):
    s = []
    for i in t:
      if not isinstance(i, (tuple, list)):
        s.append(str(i).split("'")[1])
      else:
        s.append('(' + ', '.join([str(j).split("'")[1] for j in i]) + ')')
    return ', '.join(s)

  expected, actual = to_str(expected), to_str(actual)
  ftype = 'method'
  msg = "'{}' {} ".format(fname, ftype) \
      + ("inputs", "outputs")[flag] + " ({}), but ".format(expected) \
      + ("was given", "result is")[flag] + " ({})".format(actual)
  return msg


def _compares_types(argtype, force_types):
  # True if types is satisfied the force_types
  for i, j in zip(argtype, force_types):
    if isinstance(j, (tuple, list)):
      if i not in j:
        return False
    elif i != j:
      return False
  return True


def typecheck(inputs=None, outputs=None, debug=2):
  '''Function/Method decorator. Checks decorated function's arguments are
  of the expected types.

  Parameters
  ----------
  inputs : types
      The expected types of the inputs to the decorated function.
      Must specify type for each parameter.
  outputs : types
      The expected type of the decorated function's return value.
      Must specify type for each parameter.
  debug : int, str
      Optional specification of 'debug' level:
      0:'ignore', 1:'warn', 2:'raise'

  Examples
  --------
  >>> # Function typecheck
  >>> @typecheck(inputs=(int, str, float), outputs=(str))
  >>> def function(a, b, c):
  ...     return b
  >>> function(1, '1', 1.) # no error
  >>> function(1, '1', 1) # error, final argument must be float
  ...
  >>> # method typecheck
  >>> class ClassName(object):
  ...     @typecheck(inputs=(str, int), outputs=int)
  ...     def method(self, a, b):
  ...         return b
  >>> x = ClassName()
  >>> x.method('1', 1) # no error
  >>> x.method(1, '1') # error

  '''
  if inspect.ismethod(inputs) or inspect.isfunction(inputs):
    raise ValueError('You must specify either [inputs] types or [outputs]'
                     ' types arguments.')
  # ====== parse debug ====== #
  if isinstance(debug, str):
    debug_str = debug.lower()
    if 'raise' in debug_str:
      debug = 2
    elif 'warn' in debug_str:
      debug = 1
    else:
      debug = 0
  elif debug not in (0, 1, 2):
    debug = 2
  # ====== check types ====== #
  if inputs is not None and not isinstance(inputs, (tuple, list)):
    inputs = (inputs,)
  if outputs is not None and not isinstance(outputs, (tuple, list)):
    outputs = (outputs,)

  def wrap_function(func):
    # ====== fetch arguments order ====== #
    sign = inspect.signature(func)
    args_name = []
    args_defaults = OrderedDict()
    for n, p in sign.parameters.items():
      if p.kind in (inspect.Parameter.VAR_POSITIONAL,
                    inspect.Parameter.VAR_KEYWORD):
        continue
      args_name.append(n)
      if p.default != inspect.Parameter.empty:
        args_defaults[n] = p.default

    @wraps(func)
    def wrapper(*args, **kwargs):
      input_args = list(args)
      excluded = {i: j for i, j in zip(args_name, input_args)}
      # check default kwargs
      for i, j in args_defaults.items():
        if i in excluded: # already input as positional argument
          continue
        if i in kwargs: # specified value
          input_args.append(kwargs[i])
        else: # default value
          input_args.append(j)
      ### main logic
      if debug is 0: # ignore
        return func(*args, **kwargs)
      ### Check inputs
      if inputs is not None:
        # main logic
        length = int(min(len(input_args), len(inputs)))
        argtypes = tuple(map(type, input_args))
        # TODO: smarter way to check argtypes for methods
        if not _compares_types(argtypes[:length], inputs[:length]) and\
            not _compares_types(argtypes[1:length + 1], inputs[:length]): # wrong types
          msg = _info(func.__name__, inputs, argtypes, 0)
          if debug is 1:
            print('TypeWarning:', msg)
          elif debug is 2:
            raise TypeError(msg)
      ### get results
      results = func(*args, **kwargs)
      ### Check outputs
      if outputs is not None:
        res_types = ((type(results),)
                     if not isinstance(results, (tuple, list))
                     else tuple(map(type, results)))
        length = min(len(res_types), len(outputs))
        if len(outputs) > len(res_types) or \
            not _compares_types(res_types[:length], outputs[:length]):
          msg = _info(func.__name__, outputs, res_types, 1)
          if debug is 1:
            print('TypeWarning: ', msg)
          elif debug is 2:
            raise TypeError(msg)
      ### finally everything ok
      return results
    return wrapper
  return wrap_function


# ===========================================================================
# Auto set attributes
# ===========================================================================
def autoattr(*args, **kwargs):
  '''
  Example
  -------
  >>> class ClassName(object):
  ..... def __init__(self):
  ......... super(ClassName, self).__init__()
  ......... self.arg1 = 1
  ......... self.arg2 = False
  ...... @autoattr('arg1', arg1=lambda x: x + 1)
  ...... def test1(self):
  ......... print(self.arg1)
  ...... @autoattr('arg2')
  ...... def test2(self):
  ......... print(self.arg2)
  >>> c = ClassName()
  >>> c.test1() # arg1 = 2
  >>> c.test2() # arg2 = True

  '''
  if len(args) > 0 and (inspect.ismethod(args[0]) or inspect.isfunction(args[0])):
    raise ValueError('You must specify at least 1 *args or **kwargs, all '
                     'attributes in *args will be setted to True, likewise, '
                     'all attributes in **kwargs will be setted to given '
                     'value.')
  attrs = {i: True for i in args}
  attrs.update(kwargs)

  def wrap_function(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
      results = func(*args, **kwargs)
      if len(args) > 0:
        for i, j in attrs.items():
          if hasattr(args[0], i):
            if hasattr(j, '__call__'):
              setattr(args[0], str(i), j(getattr(args[0], i)))
            else:
              setattr(args[0], str(i), j)
      return results
    return wrapper

  return wrap_function

# ===========================================================================
# Abstract static
# ===========================================================================
class abstractstatic(staticmethod):
  __slots__ = ()

  def __init__(self, function):
    super(abstractstatic, self).__init__(function)
    function.__isabstractmethod__ = True
  __isabstractmethod__ = True

# ===========================================================================
# Python utilities
# ===========================================================================
_primitives = (bool, int, float, str,
               tuple, list, dict, type, types.ModuleType, types.FunctionType,
               type(None), type(type), np.ndarray)


def func_to_str(func):
  # conver to byte
  code = cPickle.dumps(array("B", marshal.dumps(func.__code__)),
                       protocol=cPickle.HIGHEST_PROTOCOL)
  closure = None
  if func.__closure__ is not None:
    print("[WARNING] function: %s contains closure, which cannot be "
          "serialized." % str(func))
    closure = tuple([c.cell_contents for c in func.__closure__])
  defaults = func.__defaults__
  return (code, closure, defaults)


def str_to_func(s, sandbox=None):
  if isinstance(s, (tuple, list)):
    code, closure, defaults = s
  elif isinstance(s, string_types): # path to file
    if os.path.isfile(s):
      with open(s, 'rb') as f:
        code, closure, defaults = cPickle.load(f)
    else: # pickled string
      code, closure, defaults = cPickle.loads(s)
  else:
    raise ValueError("Unsupport str_to_func for type:%s" % type(s))
  code = marshal.loads(cPickle.loads(code).tobytes())
  func = types.FunctionType(code=code, name=code.co_name,
              globals=sandbox if isinstance(sandbox, Mapping) else globals(),
              closure=closure, argdefs=defaults)
  return func


def _serialize_function_sandbox(function, source):
  '''environment, dictionary (e.g. globals(), locals())
  Parameters
  ----------
  source : str
      source code of the function

  Returns
  -------
  dictionary : cPickle dumps-able dictionary to store as text
  '''
  import re
  sys_module = re.compile(r"__\w+__")

  environment = function.__globals__
  func_module = function.__module__
  sandbox = OrderedDict()
  # ====== serialize primitive type ====== #
  seen_main_function = False
  for name, val in environment.items():
    typ = None
    # ignore system modules
    if sys_module.match(name) is not None:
      continue
    # support primitive type
    if builtins.any(isinstance(val, i) for i in _primitives):
      typ = type(val)
      if isinstance(val, np.ndarray):
        val = (val.tostring(), val.dtype)
        typ = 'ndarray'
      # special case: import module
      elif isinstance(val, types.ModuleType):
        val = val.__name__
        typ = 'module'
      # edward distribution
      elif isinstance(val, type) and str(val.__module__) == 'abc' and \
      str(type(val).__module__) == "tensorflow.contrib.distributions.python.ops.distribution":
        val = val.__name__
        typ = 'edward_distribution'
      # the FunctionType itself cannot be pickled (weird!)
      elif val is types.FunctionType:
        val = None
        typ = 'function_type'
      # for some reason, pickle cannot serialize None type
      elif val is None:
        val = None
        typ = 'None'
      elif isinstance(val, Mapping):
        val = cPickle.dumps(val, protocol=cPickle.HIGHEST_PROTOCOL)
        typ = 'Mapping'
      elif inspect.isfunction(val): # special case: function
        # function might nested, so cannot find it in globals()
        if val == function:
          seen_main_function = True
        # imported function
        _ = '_main' if function == val else ''
        if val.__module__ != func_module:
          typ = 'imported_function'
          val = (val.__name__, val.__module__)
        # defined function in the same script file
        else:
          typ = 'defined_function'
          val = func_to_str(val)
        typ += _
    # finally add to sandbox valid type
    if typ is not None:
      sandbox[name] = (typ, val)
  # ====== not seen the main function ====== #
  if not seen_main_function: # mark the main function with "_main"
    sandbox['random_name_1234'] = ('defined_function_main',
                                   func_to_str(function))
  return sandbox


def _deserialize_function_sandbox(sandbox):
  '''
  environment : dictionary
      create by `serialize_sandbox`
  '''
  import marshal
  with warnings.catch_warnings():
    warnings.filterwarnings(action='ignore', category=ImportWarning)
    import importlib

  environment = {}
  defined_function = []
  main_func = None
  # first pass we deserialize all type except function type
  for name, (typ, val) in sandbox.items():
    if isinstance(typ, string_types):
      if typ == 'None':
        val = None
      elif typ == 'edward_distribution':
        try:
          import edward
          val = getattr(edward.models, val)
        except ImportError:
          raise ImportError("Cannot import 'edward' library to deserialize "
                            "the function.")
        # exec("from edward.models import %s as %s" % (val, name))
      elif typ == 'function_type':
        val = types.FunctionType
      elif typ == 'Mapping':
        val = cPickle.loads(val)
      elif typ == 'ndarray':
        val = np.fromstring(val[0], dtype=val[1])
      elif typ == 'module':
        val = importlib.import_module(val)
      elif 'imported_function' == typ:
        val = getattr(importlib.import_module(val[1]), val[0])
        if '_main' in typ: main_func = val
      elif 'defined_function' in typ:
        val = str_to_func(val, globals())
        if '_main' in typ: main_func = val
        defined_function.append(name)
    elif builtins.any(isinstance(typ, i) for i in _primitives):
      pass
    else:
      raise ValueError('Unsupport deserializing type: {}, '
                       'value: {}'.format(typ, val))
    environment[name] = val
  # ====== create all defined function ====== #
  # second pass, function all funciton and set it globales to new environment
  for name in defined_function:
    func = environment[name]
    func.__globals__.update(environment)
  return main_func, environment


class _ArgPlaceHolder_(object):
  pass


class functionable(object):

  """ Class handles save and load a function with its arguments

  This function does perfectly for following cases:
      - Pickling `lambda` function without external dependencies.
      - Pickling top-level function.
      - Pickling imported function.

  Parameters
  ----------
  func: function
      lambda or function
  arg: list
      stored arguments list for given function
  kwargs: dict
      stored keyword arguments for given function

  Note
  ----
   - Please use this function with care, only primitive variables
   are stored in pickling the function.
   - Avoid involving closure in creating function (because closure cannot
   be serialized with any mean), for example:

  Example
  -------
  >>> # Wrong way:
  >>> lambda: obj.y
  >>> # Good way (explicitly store the obj in default arguments):
  >>> lambda x=obj: x.y
  """

  def __init__(self, func, *args, **kwargs):
    super(functionable, self).__init__()
    self._function = func
    self.__name__ = self._function.__name__
    try: # sometime cannot get the source
      self._source = inspect.getsource(self._function)
    except Exception as e:
      print("[WARNING] Cannot get source code of function:", func,
            "(error:%s)" % str(e))
      self._source = None
    # try to pickle the function directly
    try:
      self._sandbox = cPickle.dumps(self._function,
          protocol=cPickle.HIGHEST_PROTOCOL)
    except Exception:
      self._sandbox = _serialize_function_sandbox(func, self._source)
    # ====== store argsmap ====== #
    sign = inspect.signature(func)
    argsmap = OrderedDict()
    for i, (n, p) in enumerate(sign.parameters.items()):
      if p.kind in (inspect.Parameter.VAR_POSITIONAL,
                    inspect.Parameter.VAR_KEYWORD):
        continue
      if i < len(args):
        argsmap[n] = args[i]
      elif n in kwargs:
        argsmap[n] = kwargs[n]
      elif p.default != inspect.Parameter.empty:
        argsmap[n] = p.default
      else:
        argsmap[n] = _ArgPlaceHolder_()
    self._argsmap = argsmap

  # ==================== Pickling methods ==================== #
  def __getstate__(self):
    # conver to byte
    return (self._sandbox,
            self._source,
            self._argsmap)

  def __setstate__(self, states):
    (self._sandbox,
     self._source,
     self._argsmap) = states
    # ====== deserialize the function ====== #
    if isinstance(self._sandbox, string_types):
      self._function = cPickle.loads(self._sandbox)
    else:
      self._function, sandbox = _deserialize_function_sandbox(self._sandbox)
    if self._function is None:
      raise RuntimeError('[funtionable] Cannot find function in sandbox.')

  # ==================== properties ==================== #
  @property
  def function(self):
    return self._function

  @property
  def name(self):
    return self._function.__name__

  @property
  def source(self):
    return self._source

  @property
  def sandbox(self):
    return self._sandbox

  # ==================== methods ==================== #
  def __call__(self, *args, **kwargs):
    final_args = self._argsmap.copy()
    for i, j in zip(final_args.keys(), args):
      final_args[i] = j
    final_args.update(kwargs)
    final_args = {i: j for i, j in final_args.items()
                  if not isinstance(j, _ArgPlaceHolder_)}
    return self._function(**final_args)

  def __str__(self):
    s = 'Name:   %s\n' % self._function.__name__
    s += 'kwargs: %s\n' % str(self._argsmap)
    if isinstance(self._sandbox, string_types):
      s += 'Sandbox: pickle-able\n'
    else:
      s += 'Sandbox:%s\n' % str(len(self._sandbox))
    s += str(self._source)
    return s[:-1]

  def __eq__(self, other):
    if self._function == other._function and \
       self._argsmap == other._argsmap:
      return True
    return False

  # ==================== update kwargs ==================== #
  def __setitem__(self, key, value):
    if not isinstance(key, (str, int, float)):
      raise ValueError('Only accept string for kwargs key or int for '
                       'index of args, but type(key)={}'.format(type(key)))
    if isinstance(key, str):
      if key in self._argsmap:
        self._argsmap[key] = value
    else:
      key = int(key)
      if key < len(self._argsmap):
        key = self._argsmap.keys()[key]
        self._argsmap[key] = value

  def __getitem__(self, key):
    if not isinstance(key, (str, int, float)):
      raise ValueError('Only accept string for kwargs key or int for '
                       'index of args, but type(key)={}'.format(type(key)))
    if isinstance(key, str):
      return self._argsmap[key]
    return self._argsmap(int(key))


# ===========================================================================
# Singleton metaclass
# ===========================================================================
def singleton(cls):
  ''' Singleton for class instance, all __init__ with same arguments return
  same instance
  @NOTE: this is copy from six.add_metaclass
  '''
  if not isinstance(cls, type):
    raise Exception('singleton decorator only accept class (type).')
  orig_vars = cls.__dict__.copy()
  slots = orig_vars.get('__slots__')
  if slots is not None:
    if isinstance(slots, str):
      slots = [slots]
    for slots_var in slots:
      orig_vars.pop(slots_var)
  orig_vars.pop('__dict__', None)
  orig_vars.pop('__weakref__', None)
  return Singleton(cls.__name__, cls.__bases__, orig_vars)


class Singleton(type):
  # class_type -> [(arguments, instance), ...]
  _INSTANCES = defaultdict(list)

  @staticmethod
  def _dispose(self):
    clz = self.__class__
    Singleton._INSTANCES[clz] = [(args, obj)
                                 for args, obj in Singleton._INSTANCES[clz]
                                 if obj != self]

  def __new__(mcs, name, bases, class_dict):
    if '_get_id' not in class_dict:
      raise ValueError("Instance of Singleton must define classmethod "
          "'_get_id', this method takes the same arguments as __init__ "
          "and return the unique identity for an instance.")
    return super().__new__(mcs, name, bases, class_dict)

  def __call__(cls, *args, **kwargs):
    obj_id = cls._get_id(*args, **kwargs)
    # check defined instance
    instances_list = Singleton._INSTANCES[cls]
    for arguments, obj in instances_list:
      if arguments == obj_id:
        return obj
    # Create new instance
    obj = super(Singleton, cls).__call__(*args, **kwargs)
    instances_list.append((obj_id, obj))
    setattr(obj, '__del__',
            types.MethodType(Singleton._dispose, obj))
    return obj



from __future__ import absolute_import, division, print_function

import numpy as np
from tensorflow.python.keras.callbacks import Callback
from tensorflow.python.platform import tf_logging as logging


class EarlyStopping(Callback):
  """ Original implementation from keras, copied here with some
  improvement.

  Stop training when a monitored quantity has stopped improving.

  Arguments:
      monitor: Quantity to be monitored.
      min_delta: Minimum change in the monitored quantity
          to qualify as an improvement, i.e. an absolute
          change of less than min_delta, will count as no
          improvement.
      patience: Number of epochs with no improvement
          after which training will be stopped.
      verbose: verbosity mode.
      mode: One of `{"auto", "min", "max"}`. In `min` mode,
          training will stop when the quantity
          monitored has stopped decreasing; in `max`
          mode it will stop when the quantity
          monitored has stopped increasing; in `auto`
          mode, the direction is automatically inferred
          from the name of the monitored quantity.
      baseline: Baseline value for the monitored quantity.
          Training will stop if the model doesn't show improvement over the
          baseline.
      restore_best_weights: Whether to restore model weights from
          the epoch with the best value of the monitored quantity.
          If False, the model weights obtained at the last step of
          training are used.

  Example:

  ```python
  callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=3)
  # This callback will stop the training when there is no improvement in
  # the validation loss for three consecutive epochs.
  model.fit(data, labels, epochs=100, callbacks=[callback],
      validation_data=(val_data, val_labels))
  ```
  """

  def __init__(self,
               monitor='val_loss',
               min_delta=0,
               patience=0,
               verbose=0,
               mode='auto',
               baseline=None,
               terminate_on_nan=True,
               restore_best_weights=False):
    super(EarlyStopping, self).__init__()

    self.monitor = monitor
    self.patience = patience
    self.verbose = verbose
    self.baseline = baseline
    self.min_delta = abs(min_delta)
    self.wait = 0
    self.stopped_epoch = 0
    self.terminate_on_nan = bool(terminate_on_nan)
    self.restore_best_weights = restore_best_weights
    self.best_weights = None
    self.best_epoch = None

    if mode not in ['auto', 'min', 'max']:
      logging.warning(
          'EarlyStopping mode %s is unknown, '
          'fallback to auto mode.', mode)
      mode = 'auto'

    if mode == 'min':
      self.monitor_op = np.less
    elif mode == 'max':
      self.monitor_op = np.greater
    else:
      if 'acc' in self.monitor:
        self.monitor_op = np.greater
      else:
        self.monitor_op = np.less

    if self.monitor_op == np.greater:
      self.min_delta *= 1
    else:
      self.min_delta *= -1

  def on_train_begin(self, logs=None):
    # Allow instances to be re-used
    self.wait = 0
    self.stopped_epoch = 0
    if self.baseline is not None:
      self.best = self.baseline
    else:
      self.best = np.Inf if self.monitor_op == np.less else -np.Inf

  def on_batch_end(self, batch, logs=None):
    logs = logs or {}
    loss = logs.get('loss')
    if loss is not None:
      if np.isnan(loss) or np.isinf(loss):
        # terminate on NaN
        if self.terminate_on_nan:
          print('Batch %d: Invalid loss, terminating training' % (batch))
          self.model.stop_training = True
        # restoring the best weights
        if self.restore_best_weights and self.best_weights is not None:
          if self.verbose > 0:
            print(
                'Restoring model weights from the end of the best epoch #%d.' %
                self.best_epoch)
          self.model.set_weights(self.best_weights)

  def on_epoch_end(self, epoch, logs=None):
    current = self.get_monitor_value(logs)
    if current is None:
      return
    if self.monitor_op(current - self.min_delta, self.best):
      self.best = current
      self.wait = 0
      if self.restore_best_weights:
        self.best_weights = self.model.get_weights()
        self.best_epoch = epoch
    else:
      self.wait += 1
      if self.wait >= self.patience:
        self.stopped_epoch = epoch
        self.model.stop_training = True
        if self.restore_best_weights:
          if self.verbose > 0:
            print(
                'Restoring model weights from the end of the best epoch #%d.' %
                self.best_epoch)
          self.model.set_weights(self.best_weights)

  def on_train_end(self, logs=None):
    if self.stopped_epoch > 0 and self.verbose > 0:
      print('Epoch %05d: early stopping' % (self.stopped_epoch + 1))

  def get_monitor_value(self, logs):
    logs = logs or {}
    monitor_value = logs.get(self.monitor)
    if monitor_value is None:
      logging.warning(
          'Early stopping conditioned on metric `%s` '
          'which is not available. Available metrics are: %s', self.monitor,
          ','.join(list(logs.keys())))
    return monitor_value



from __future__ import absolute_import, division, print_function

import inspect
import io
import numbers
import os
import re
import string
import tarfile
import types
import warnings
from collections import (Iterable, Iterator, Mapping, OrderedDict, defaultdict,
                         deque)
from contextlib import contextmanager
from datetime import datetime

import numpy as np
from six import add_metaclass, string_types
from six.moves import cPickle

# ===========================================================================
# File type check
# ===========================================================================
GZIP_MAGIC_NUMBER = "1f8b"


def is_gzip_file(path):
  """ Credit:
  https://kite.com/python/examples/4945/gzip-check-if-a-file-is-gzip-compressed
  """
  if isinstance(path, string_types) and os.path.isfile(path):
    with open(path, 'rb') as f:
      return f.read(2).encode("hex") == GZIP_MAGIC_NUMBER
  elif hasattr(path, 'read') and hasattr(path, 'tell'):
    last_pos = path.tell()
    path.seek(0)
    indicator = path.read(2)
    indicator = (indicator.encode("hex")
                 if isinstance(indicator, string_types) else indicator.hex())
    path.seek(last_pos)
    return indicator == GZIP_MAGIC_NUMBER
  return False


def is_tar_file(path):
  if not os.path.isfile(path):
    return False
  return tarfile.is_tarfile(path)


# ===========================================================================
# Regular expression
# ===========================================================================
RE_NUMBER = re.compile(r'^[+-]*((\d*\.\d+)|(\d+))$')


# ===========================================================================
# Getter
# ===========================================================================
def get_formatted_datetime(only_number=True):
  if only_number:
    return "{:%H%M%S%d%m%y}".format(datetime.now())
  return "{:%H:%M:%S-%d%b%y}".format(datetime.now())


def get_all_properties(obj):
  """ Return all attributes which are properties of given Object
  """
  properties = []
  clazz = obj if isinstance(obj, type) else obj.__class__
  for key in dir(clazz):
    if '__' in key:
      continue
    val = getattr(clazz, key)
    if isinstance(val, property):
      properties.append(key)
  return properties if isinstance(obj, type) else \
  {p: getattr(obj, p) for p in properties}


def get_string_placeholders(s):
  assert isinstance(s, string_types)
  fmt = []
  for (_, key, spec, _) in string.Formatter().parse(save_path):
    if spec is not None:
      fmt.append(key)
  return tuple(fmt)


# ===========================================================================
# Data converter
# ===========================================================================
def as_tuple(x, N=None, t=None):
  """
  Coerce a value to a tuple of given length (and possibly given type).

  Parameters
  ----------
  x : {value, iterable}
  N : {integer}
      length of the desired tuple
  t : {type, call-able, optional}
      required type for all elements

  Returns
  -------
  tuple
      ``tuple(x)`` if `x` is iterable, ``(x,) * N`` otherwise.

  Raises
  ------
  TypeError
      if `type` is given and `x` or any of its elements do not match it
  ValueError
      if `x` is iterable, but does not have exactly `N` elements

  Note
  ----
  This function is adpated from Lasagne
  Original work Copyright (c) 2014-2015 lasagne contributors
  All rights reserved.

  LICENSE: https://github.com/Lasagne/Lasagne/blob/master/LICENSE
  """
  # special case numpy array
  if not isinstance(x, tuple):
    if isinstance(x, (types.GeneratorType, list)):
      x = tuple(x)
    else:
      x = (x,)
  # ====== check length ====== #
  if is_number(N):
    N = int(N)
    if len(x) == 1:
      x = x * N
    elif len(x) != N:
      raise ValueError('x has length=%d, but required length N=%d' %
                       (len(x), N))
  # ====== check type ====== #
  if t is None:
    filter_func = lambda o: True
  elif isinstance(t, type) or isinstance(t, (tuple, list)):
    filter_func = lambda o: isinstance(o, t)
  elif hasattr(t, '__call__'):
    filter_func = t
  else:
    raise ValueError("Invalid value for `t`: %s" % str(t))
  if not all(filter_func(v) for v in x):
    raise TypeError("expected a single value or an iterable "
                    "of {0}, got {1} instead".format(t.__name__, x))
  return x


def as_list(x, N=None, t=None):
  return list(as_tuple(x, N, t))


def as_bytes(x, nbytes=None, order='little'):
  """ Convert some python object to bytes array, support type:
  * string, unicode
  * integer
  * numpy.ndarray

  Note
  ----
  This method is SLOW
  """
  if is_string(x):
    return x.encode()
  elif isinstance(x, int):
    return x.to_bytes(nbytes, order, signed=False)
  elif isinstance(x, np.ndarray):
    return x.tobytes()
  else:
    raise ValueError("Not support bytes conversion for type: %s" %
                     type(x).__name__)


# ===========================================================================
# Types check
# ===========================================================================
def is_lambda(v):
  LAMBDA = lambda: 0
  return isinstance(v, type(LAMBDA)) and v.__name__ == LAMBDA.__name__


def is_pickleable(x):
  try:
    cPickle.dumps(x, protocol=cPickle.HIGHEST_PROTOCOL)
    return True
  except cPickle.PickleError:
    return False


def is_fileobj(f):
  """ Check if an object `f` is intance of FileIO object created
  by `open()`"""
  return isinstance(f, io.TextIOBase) or \
      isinstance(f, io.BufferedIOBase) or \
      isinstance(f, io.RawIOBase) or \
      isinstance(f, io.IOBase)


def is_callable(x):
  return hasattr(x, '__call__')


def is_string(s):
  return isinstance(s, string_types)


def is_path(path):
  if is_string(path):
    try:
      os.path.exists(path)
      return True
    except Exception as e:
      return False
  return False


def is_number(i, string_number=False):
  if isinstance(i, string_types) and string_number:
    return RE_NUMBER.match(i) is not None
  return isinstance(i, numbers.Number)


def is_bool(b):
  return isinstance(b, type(True))


def is_primitives(x, inc_ndarray=True, exception_types=[]):
  """Primitive types include: number, string, boolean, None
  and numpy.ndarray (optional) and numpy.generic (optional)

  Parameters
  ----------
  inc_ndarray: bool
      if True, include `numpy.ndarray` and `numpy.generic` as a primitive types
  """
  # complex list or Mapping
  if isinstance(x, (tuple, list)):
    return all(
        is_primitives(
            i, inc_ndarray=inc_ndarray, exception_types=exception_types)
        for i in x)
  elif isinstance(x, Mapping):
    return all(
        is_primitives(
            i, inc_ndarray=inc_ndarray, exception_types=exception_types) and
        is_primitives(
            j, inc_ndarray=inc_ndarray, exception_types=exception_types)
        for i, j in x.items())
  # check for number, string, bool, and numpy array
  if is_number(x) or is_string(x) or is_bool(x) or x is None or \
  (any(isinstance(x, t) for t in exception_types)) or \
  (inc_ndarray and isinstance(x, (np.ndarray, np.generic))):
    return True
  return False


# ===========================================================================
# IO utilities
# ===========================================================================
def savetxt(fname,
            X,
            fmt='%g',
            delimiter=' ',
            newline='\n',
            header='',
            footer='',
            index=None,
            comments='# ',
            encoding=None,
            async_backend='thread'):
  """ Save an array to a text file.

  Parameters
  ----------

  fname : filename or file handle
      If the filename ends in .gz, the file is automatically saved
      in compressed gzip format. loadtxt understands gzipped files
      transparently.

  X : 1D or 2D array_like
      Data to be saved to a text file.

  fmt : str or sequence of strs, optional
      A single format (%10.5f), a sequence of formats, or a multi-format
      string, e.g. ‘Iteration %d – %10.5f’, in which case delimiter is
      ignored. For complex X, the legal options for fmt are:

      a single specifier, fmt=’%.4e’, resulting in numbers formatted like
      ‘ (%s+%sj)’ % (fmt, fmt)
      a full string specifying every real and imaginary part, e.g.
      ‘ %.4e %+.4ej %.4e %+.4ej %.4e %+.4ej’ for 3 columns
      a list of specifiers, one per column - in this case, the real and
      imaginary part must have separate specifiers,
      e.g. [‘%.3e + %.3ej’, ‘(%.15e%+.15ej)’] for 2 columns

  delimiter : str, optional
      String or character separating columns.

  newline : str, optional
      String or character separating lines.

  header : str, optional
      String that will be written at the beginning of the file.

  footer : str, optional
      String that will be written at the end of the file.

  comments : str, optional
      String that will be prepended to the header and footer strings, to mark them as comments. Default: ‘# ‘, as expected by e.g. numpy.loadtxt.

  encoding : {None, str}, optional
      Encoding used to encode the outputfile. Does not apply to output streams. If the encoding is something other than ‘bytes’ or ‘latin1’ you will not be able to load the file in NumPy versions < 1.14. Default is ‘latin1’.

  async_backend : {'thread', 'process', None}
      save the data to disk asynchronously using multi-threading or
      multi-processing
  """
  pass


# ===========================================================================
# String processing
# ===========================================================================
_space_char = re.compile(r"\s")
_multiple_spaces = re.compile(r"\s\s+")
_non_alphanumeric_char = re.compile(r"\W")


def string_normalize(text,
                     lower=True,
                     remove_non_alphanumeric=True,
                     remove_duplicated_spaces=True,
                     remove_whitespace=False,
                     escape_pattern=False):
  text = str(text).strip()
  if bool(lower):
    text = text.lower()
  if bool(escape_pattern):
    text = re.escape(text)
  if bool(remove_non_alphanumeric):
    text = _non_alphanumeric_char.sub(' ', text)
    text = text.strip()
  if bool(remove_duplicated_spaces):
    text = _multiple_spaces.sub(' ', text)
  if bool(remove_whitespace):
    if isinstance(remove_whitespace, string_types):
      text = _space_char.sub(remove_whitespace, text)
    else:
      text = _space_char.sub('', text)
  return text


text_normalize = string_normalize


# ===========================================================================
# List utils
# ===========================================================================
def unique(seq, keep_order=False):
  if keep_order:
    seen = set()
    seen_add = seen.add
    return [x for x in seq if x not in seen and not seen_add(x)]
  else:
    return list(set(seq))


# ===========================================================================
# Async file IO
# ===========================================================================
class defaultdictkey(defaultdict):
  """ Enhanced version of `defaultdict`, instead of return a
  default value, return an "improvised" default value based on
  the given key.

  Example
  -------
  >>> from odin.utils.python_utils import defaultdictkey
  >>> d = defaultdictkey(lambda x: str(x))
  >>> print(d['123']) # '123'
  """

  def __missing__(self, key):
    if self.default_factory is None:
      raise KeyError(key)
    else:
      ret = self[key] = self.default_factory(key)
      return ret


def multikeysdict(d):
  assert isinstance(d, dict)
  new_d = d.__class__()
  for i, j in d.items():
    if isinstance(i, tuple):
      for k in i:
        new_d[k] = j
    else:
      new_d[i] = j
  return new_d


# ===========================================================================
# Object and class attributes
# ===========================================================================
class abstractclassmethod(classmethod):

  __isabstractmethod__ = True

  def __init__(self, method):
    method.__isabstractmethod__ = True
    super(abstractclassmethod, self).__init__(method)


class classproperty(object):

  def __init__(self, fn):
    super(classproperty, self).__init__()
    self.fn = fn

  def __get__(self, obj, owner):
    return self.fn(owner)


# ===========================================================================
# Path utils
# ===========================================================================
def select_path(*paths, default=None, create_new=False):
  """
  Parameters
  ----------
  paths : str
    multiple path are given

  default : str
    default path for return

  create_new : bool (default: False)
    if no path is found, create new folder based on the
    first path found to be `creat-able`
  """
  all_paths = []
  for p in paths:
    if isinstance(p, (tuple, list)):
      all_paths += p
    elif isinstance(p, string_types):
      all_paths.append(p)
    else:
      raise ValueError("Given `path` has type: '%s', which must be string or "
                       "list of string")
  # ====== return the first found exists path ====== #
  for p in all_paths:
    if os.path.exists(p):
      return p
  # ====== check if create_new ====== #
  if default is not None:
    return str(default)
  if create_new:
    for p in paths:
      base_dir = os.path.dirname(p)
      if os.path.exists(base_dir):
        os.mkdir(p)
        return p
    raise ValueError("Cannot create new folder from list: %s" % str(paths))
  # ====== raise exception ====== #
  raise RuntimeError("Cannot find any exists path from list: %s" %
                     '; '.join(all_paths))


# ===========================================================================
# Warnings and Exception
# ===========================================================================
@contextmanager
def catch_warnings_error(w):
  """ This method turn any given warnings into exception

  use: `warnings.Warning` for all warnings

  Example
  -------
  >>> with catch_warnings([RuntimeWarning, UserWarning]):
  >>>   try:
  >>>     warnings.warn('test', category=RuntimeWarning)
  >>>   except RuntimeWarning as w:
  >>>     pass
  """
  with warnings.catch_warnings():
    warnings.filterwarnings(action='error', category=w)
    yield


@contextmanager
def catch_warnings_ignore(w):
  """ This method ignore any given warnings

  use: `warnings.Warning` for all warnings
  """
  with warnings.catch_warnings():
    warnings.filterwarnings(action='ignore', category=w)
    yield



import os
import pickle

import numpy as np
from scipy.io import savemat

from odin import fuel as F, visual as V
from odin.utils import ctext, Progbar, get_exppath, select_path
from odin.stats import train_valid_test_split, sampling_iter
from odin.preprocessing.signal import segment_axis

HOME_PATH = os.path.expanduser('~')
# fixed path to 'voxceleb1_wav' folder
PATH_TO_WAV = select_path(
    '/media/data2/SRE_DATA/voxceleb',
    '/mnt/sdb1/SRE_DATA/voxceleb',
    os.path.join(HOME_PATH, 'data', 'voxceleb'),
    os.path.join(HOME_PATH, 'voxceleb'),
    create_new=False
)
# path to folder contains experiment results
PATH_EXP = get_exppath('voxceleb')
# output path for acoustic features directory
PATH_ACOUSTIC_FEAT = os.path.join(PATH_EXP, 'voxceleb_feat')
if not os.path.exists(PATH_ACOUSTIC_FEAT):
  os.mkdir(PATH_ACOUSTIC_FEAT)
# ====== remove '_quarter' if you want full training data ====== #
FILE_LIST = "voxceleb_files_quarter"
TRAIN_LIST = "voxceleb_sys_train_with_labels_quarter"
TRIAL_LIST = "voxceleb_trials"
# ====== Load the file list ====== #
ds = F.load_voxceleb_list()
WAV_FILES = {} # dictionary mapping 'file_path' -> 'file_name'
for path, channel, name in ds[FILE_LIST]:
  path = os.path.join(PATH_TO_WAV, path)
  # validate all files are exist
  assert os.path.exists(path), path
  WAV_FILES[path] = name
# some sampled files for testing
SAMPLED_WAV_FILE = sampling_iter(it=sorted(WAV_FILES.items(),
                                           key=lambda x: x[0]),
                                 k=8, seed=87654321)
# ====== extract the list of all train files ====== #
# mapping from name of training file to speaker label
TRAIN_DATA = {}
for x, y in ds[TRAIN_LIST]:
  TRAIN_DATA[x] = int(y)

# ===========================================================================
# Path helpers
# ===========================================================================
def get_model_path(system_name, args):
  """Return: exp_dir, model_path, log_path, train_path, test_path"""
  name = '_'.join([str(system_name).lower(), args.recipe, args.feat])
  if 'l' in args:
    name += '_' + str(int(args.l))
  if 'nmix' in args:
    name += '_' + str(int(args.nmix))
  if 'tdim' in args:
    name += '_' + str(int(args.tdim))
  save_path = os.path.join(PATH_EXP, name)
  if not os.path.exists(save_path):
    os.mkdir(save_path)
  # ====== return path ====== #
  log_path = os.path.join(save_path, 'log.txt')
  model_path = os.path.join(save_path, 'model.ai')
  train_path = os.path.join(save_path, 'train.dat')
  test_path = os.path.join(save_path, 'test.dat')
  print("Model path:", ctext(model_path, 'cyan'))
  print("Log path:", ctext(log_path, 'cyan'))
  return save_path, model_path, log_path, train_path, test_path

# ===========================================================================
# Data helpers
# ===========================================================================
FRAME_SHIFT = 0.005

def prepare_dnn_data(recipe, feat, utt_length, seed=87654321):
  """
  Return
  ------
  train_feeder : Feeder for training
  valid_feeder : Feeder for validating
  test_ids : Test indices
  test_dat : Data array
  all_speakers : list of all speaker in training set
  """
  # Load dataset
  frame_length = int(utt_length / FRAME_SHIFT)
  ds = F.Dataset(os.path.join(PATH_ACOUSTIC_FEAT, recipe),
                 read_only=True)
  X = ds[feat]
  train_indices = {name: ds['indices'][name]
                   for name in TRAIN_DATA.keys()}
  test_indices = {name: start_end
                  for name, start_end in ds['indices'].items()
                  if name not in TRAIN_DATA}
  train_indices, valid_indices = train_valid_test_split(
      x=list(train_indices.items()), train=0.9, inc_test=False, seed=seed)
  all_speakers = sorted(set(TRAIN_DATA.values()))
  n_speakers = max(all_speakers) + 1
  print("#Train files:", ctext(len(train_indices), 'cyan'))
  print("#Valid files:", ctext(len(valid_indices), 'cyan'))
  print("#Test files:", ctext(len(test_indices), 'cyan'))
  print("#Speakers:", ctext(n_speakers, 'cyan'))
  recipes = [
      F.recipes.Sequencing(frame_length=frame_length, step_length=frame_length,
                           end='pad', pad_value=0, pad_mode='post',
                           data_idx=0),
      F.recipes.Name2Label(lambda name:TRAIN_DATA[name], ref_idx=0),
      F.recipes.LabelOneHot(nb_classes=n_speakers, data_idx=1)
  ]
  train_feeder = F.Feeder(
      data_desc=F.IndexedData(data=X, indices=train_indices),
      batch_mode='batch', ncpu=7, buffer_size=12)
  valid_feeder = F.Feeder(
      data_desc=F.IndexedData(data=X, indices=valid_indices),
      batch_mode='batch', ncpu=2, buffer_size=4)
  train_feeder.set_recipes(recipes)
  valid_feeder.set_recipes(recipes)
  print(train_feeder)
  # ====== cache the test data ====== #
  cache_dat = os.path.join(PATH_EXP, 'test_%s_%d.dat' % (feat, int(utt_length)))
  cache_ids = os.path.join(PATH_EXP, 'test_%s_%d.ids' % (feat, int(utt_length)))
  # validate cache files
  if os.path.exists(cache_ids):
    with open(cache_ids, 'rb') as f:
      ids = pickle.load(f)
    if len(ids) != len(test_indices):
      os.remove(cache_ids)
      if os.path.exists(cache_dat):
        os.remove(cache_dat)
  elif os.path.exists(cache_dat):
    os.remove(cache_dat)
  # caching
  if not os.path.exists(cache_dat):
    dat = F.MmapData(cache_dat, dtype='float16',
                     shape=(0, frame_length, X.shape[1]))
    ids = {}
    prog = Progbar(target=len(test_indices))
    s = 0
    for name, (start, end) in test_indices.items():
      y = X[start:end]
      y = segment_axis(y, axis=0,
                       frame_length=frame_length, step_length=frame_length,
                       end='pad', pad_value=0, pad_mode='post')
      dat.append(y)
      # update indices
      ids[name] = (s, s + len(y))
      s += len(y)
      # update progress
      prog.add(1)
    dat.flush()
    dat.close()
    with open(cache_ids, 'wb') as f:
      pickle.dump(ids, f)
  # ====== re-load ====== #
  dat = F.MmapData(cache_dat, read_only=True)
  with open(cache_ids, 'rb') as f:
    ids = pickle.load(f)
  # ====== save some sample ====== #
  sample_path = os.path.join(PATH_EXP,
                             'test_%s_%d.pdf' % (feat, int(utt_length)))
  V.plot_figure(nrow=9, ncol=6)
  for i, (name, (start, end)) in enumerate(
      sampling_iter(it=sorted(ids.items(), key=lambda x: x[0]), k=12, seed=87654321)):
    x = dat[start:end][:].astype('float32')
    ax = V.plot_spectrogram(x[np.random.randint(0, len(x))].T,
                            ax=(12, 1, i + 1), title='')
    ax.set_title(name)
  V.plot_save(sample_path)
  return (train_feeder, valid_feeder,
          ids, dat, all_speakers)

def prepare_ivec_data(recipe, feat):
  ds = F.Dataset(os.path.join(PATH_ACOUSTIC_FEAT, recipe),
                 read_only=True)
  X = ds[feat]
  train_indices = {name: ds['indices'][name]
                   for name in TRAIN_DATA.keys()}
  test_indices = {name: start_end
                  for name, start_end in ds['indices'].items()
                  if name not in TRAIN_DATA}
  print("#Train files:", ctext(len(train_indices), 'cyan'))
  print("#Test files:", ctext(len(test_indices), 'cyan'))
  return X, train_indices, test_indices

# ===========================================================================
# Data saving helpers
# ===========================================================================
def csv2mat(exp_dir):
  in_train = os.path.join(exp_dir, 'train.dat')
  out_train = os.path.join(exp_dir, 'train.mat')
  in_test = os.path.join(exp_dir, 'test.dat')
  out_test = os.path.join(exp_dir, 'test.mat')
  assert os.path.exists(in_train) and os.path.exists(in_test)

  train_dat = np.genfromtxt(in_train, dtype=str, delimiter='\t')
  X_train = train_dat[:, 1:].astype('float32').T
  y_train = train_dat[:, 0].ravel()
  print('Train:', X_train.shape, X_train.dtype, y_train.shape, y_train.dtype)
  savemat(file_name=out_train, mdict={'X': X_train.T, 'y': y_train})

  test_dat = np.genfromtxt(in_test, dtype=str, delimiter='\t')
  X_test = test_dat[:, 1:].astype('float32').T
  y_test = test_dat[:, 0].ravel()
  print('Test:', X_test.shape, X_test.dtype, y_test.shape, y_test.dtype)
  savemat(file_name=out_test, mdict={'X': X_test.T, 'y': y_test})
